import abc
import typing
from datetime import datetime
from typing import (
    Any,
    Awaitable,
    Callable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    ValuesView,
)

from mode import Seconds, ServiceT
from mode.utils.collections import FastUserDict, ManagedUserDict
from yarl import URL

from .codecs import CodecArg
from .events import EventT
from .stores import StoreT
from .streams import JoinableT
from .topics import TopicT
from .tuples import FutureMessage, TP, MessageSentCallback
from .windows import WindowT

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
    from .models import FieldDescriptorT as _FieldDescriptorT
    from .models import ModelArg as _ModelArg
    from .serializers import SchemaT as _SchemaT
else:
    class _AppT: ...  # noqa
    class _FieldDescriptorT: ...  # noqa
    class _ModelArg: ...  # noqa
    class _SchemaT: ...   # noqa

__all__ = [
    'RecoverCallback',
    'RelativeArg',
    'CollectionT',
    'TableT',
    'GlobalTableT',
    'TableManagerT',
    'WindowCloseCallback',
    'WindowSetT',
    'WindowedItemsViewT',
    'WindowedValuesViewT',
    'WindowWrapperT',
    'ChangelogEventCallback',
    'CollectionTps',
]

RelativeHandler = Callable[[Optional[EventT]], Union[float, datetime]]
RecoverCallback = Callable[[], Awaitable[None]]
ChangelogEventCallback = Callable[[EventT], Awaitable[None]]
WindowCloseCallback = Callable[[Any, Any], None]
RelativeArg = Optional[Union[
    _FieldDescriptorT,
    RelativeHandler,
    datetime,
    float,
]]
CollectionTps = MutableMapping['CollectionT', Set[TP]]

KT = TypeVar('KT')
VT = TypeVar('VT')


class CollectionT(ServiceT, JoinableT):
    app: _AppT
    name: str
    default: Any  # noqa: E704
    schema: Optional[_SchemaT]
    key_type: Optional[_ModelArg]
    value_type: Optional[_ModelArg]
    partitions: Optional[int]
    window: Optional[WindowT]
    help: str
    recovery_buffer_size: int
    standby_buffer_size: int
    options: Optional[Mapping[str, Any]]
    last_closed_window: float
    use_partitioner: bool

    is_global: bool = False

    @abc.abstractmethod
    def __init__(self,
                 app: _AppT,
                 *,
                 name: str = None,
                 default: Callable[[], Any] = None,
                 store: Union[str, URL] = None,
                 schema: _SchemaT = None,
                 key_type: _ModelArg = None,
                 value_type: _ModelArg = None,
                 partitions: int = None,
                 window: WindowT = None,
                 changelog_topic: TopicT = None,
                 help: str = None,
                 on_recover: RecoverCallback = None,
                 on_changelog_event: ChangelogEventCallback = None,
                 recovery_buffer_size: int = 1000,
                 standby_buffer_size: int = None,
                 extra_topic_configs: Mapping[str, Any] = None,
                 options: Mapping[str, Any] = None,
                 use_partitioner: bool = False,
                 on_window_close: WindowCloseCallback = None,
                 **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    def clone(self, **kwargs: Any) -> Any:
        ...

    @property
    @abc.abstractmethod
    def changelog_topic(self) -> TopicT:
        ...

    @changelog_topic.setter
    def changelog_topic(self, topic: TopicT) -> None:
        ...

    @abc.abstractmethod
    def _changelog_topic_name(self) -> str:
        ...

    @abc.abstractmethod
    def apply_changelog_batch(self, batch: Iterable[EventT]) -> None:
        ...

    @abc.abstractmethod
    def persisted_offset(self, tp: TP) -> Optional[int]:
        ...

    @abc.abstractmethod
    async def need_active_standby_for(self, tp: TP) -> bool:
        ...

    @abc.abstractmethod
    def reset_state(self) -> None:
        ...

    @abc.abstractmethod
    def send_changelog(self,
                       partition: Optional[int],
                       key: Any,
                       value: Any,
                       key_serializer: CodecArg = None,
                       value_serializer: CodecArg = None,
                       callback: MessageSentCallback = None) -> FutureMessage:
        ...

    @abc.abstractmethod
    def partition_for_key(self, key: Any) -> Optional[int]:
        ...

    @abc.abstractmethod
    def on_window_close(self, key: Any, value: Any) -> None:
        ...

    @abc.abstractmethod
    async def on_rebalance(self,
                           assigned: Set[TP],
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    async def on_changelog_event(self, event: EventT) -> None:
        ...

    @abc.abstractmethod
    def on_recover(self, fun: RecoverCallback) -> RecoverCallback:
        ...

    @abc.abstractmethod
    async def on_recovery_completed(self,
                                    active_tps: Set[TP],
                                    standby_tps: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    async def call_recover_callbacks(self) -> None:
        ...

    @abc.abstractmethod
    def using_window(self, window: WindowT, *,
                     key_index: bool = False) -> 'WindowWrapperT':
        ...

    @abc.abstractmethod
    def hopping(self, size: Seconds, step: Seconds,
                expires: Seconds = None,
                key_index: bool = False) -> 'WindowWrapperT':
        ...

    @abc.abstractmethod
    def tumbling(self, size: Seconds,
                 expires: Seconds = None,
                 key_index: bool = False) -> 'WindowWrapperT':
        ...

    @abc.abstractmethod
    def as_ansitable(self, **kwargs: Any) -> str:
        ...

    @abc.abstractmethod
    def _relative_now(self, event: EventT = None) -> float:
        ...

    @abc.abstractmethod
    def _relative_event(self, event: EventT = None) -> float:
        ...

    @abc.abstractmethod
    def _relative_field(self, field: _FieldDescriptorT) -> RelativeHandler:
        ...

    @abc.abstractmethod
    def _relative_timestamp(self, timestamp: float) -> RelativeHandler:
        ...

    @abc.abstractmethod
    def _windowed_contains(self, key: Any, timestamp: float) -> bool:
        ...

    @abc.abstractmethod
    def prefix_scan(self, prefix: Any) -> Iterator[Tuple[Any, Any]]:
        ...        


class TableT(CollectionT, ManagedUserDict[KT, VT]):
    ...


class GlobalTableT(TableT):
    ...


class TableManagerT(ServiceT, FastUserDict[str, CollectionT]):
    app: _AppT

    actives_ready: bool
    standbys_ready: bool

    @abc.abstractmethod
    def __init__(self, app: _AppT, **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    def add(self, table: CollectionT) -> CollectionT:
        ...

    @abc.abstractmethod
    def persist_offset_on_commit(self,
                                 store: StoreT,
                                 tp: TP,
                                 offset: int) -> None:
        ...

    @abc.abstractmethod
    def on_commit(self, offsets: MutableMapping[TP, int]) -> None:
        ...

    @abc.abstractmethod
    async def on_rebalance(self,
                           assigned: Set[TP],
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    def on_partitions_revoked(self, revoked: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    def on_rebalance_start(self) -> None:
        ...

    @abc.abstractmethod
    async def wait_until_tables_registered(self) -> None:
        ...

    @abc.abstractmethod
    async def wait_until_recovery_completed(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def changelog_topics(self) -> Set[str]:
        ...


class WindowSetT(FastUserDict[KT, VT]):
    key: Any
    table: TableT
    event: Optional[EventT]

    @abc.abstractmethod
    def __init__(self,
                 key: KT,
                 table: TableT,
                 wrapper: 'WindowWrapperT',
                 event: EventT = None) -> None:
        ...

    @abc.abstractmethod
    def apply(self,
              op: Callable[[VT, VT], VT],
              value: VT,
              event: EventT = None) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def value(self, event: EventT = None) -> VT:
        ...

    @abc.abstractmethod
    def current(self, event: EventT = None) -> VT:
        ...

    @abc.abstractmethod
    def now(self) -> VT:
        ...

    @abc.abstractmethod
    def delta(self, d: Seconds, event: EventT = None) -> VT:
        ...

    @abc.abstractmethod
    def __iadd__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __isub__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __imul__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __itruediv__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __ifloordiv__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __imod__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __ipow__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __ilshift__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __irshift__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __iand__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __ixor__(self, other: VT) -> 'WindowSetT':
        ...

    @abc.abstractmethod
    def __ior__(self, other: VT) -> 'WindowSetT':
        ...


class WindowedItemsViewT(ItemsView):

    @abc.abstractmethod
    def __init__(self,
                 mapping: 'WindowWrapperT',
                 event: EventT = None) -> None:
        ...

    @abc.abstractmethod
    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        ...

    @abc.abstractmethod
    def now(self) -> Iterator[Tuple[Any, Any]]:
        ...

    @abc.abstractmethod
    def current(self, event: EventT = None) -> Iterator[Tuple[Any, Any]]:
        ...

    @abc.abstractmethod
    def delta(self,
              d: Seconds,
              event: EventT = None) -> Iterator[Tuple[Any, Any]]:
        ...


class WindowedValuesViewT(ValuesView):

    @abc.abstractmethod
    def __init__(self,
                 mapping: 'WindowWrapperT',
                 event: EventT = None) -> None:
        ...

    @abc.abstractmethod
    def __iter__(self) -> Iterator[Any]:
        ...

    @abc.abstractmethod
    def now(self) -> Iterator[Any]:
        ...

    @abc.abstractmethod
    def current(self, event: EventT = None) -> Iterator[Any]:
        ...

    @abc.abstractmethod
    def delta(self, d: Seconds, event: EventT = None) -> Iterator[Any]:
        ...


class WindowWrapperT(MutableMapping):
    table: TableT

    @abc.abstractmethod
    def __init__(self, table: TableT, *,
                 relative_to: RelativeArg = None,
                 key_index: bool = False,
                 key_index_table: TableT = None) -> None:
        ...

    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...

    @abc.abstractmethod
    def clone(self, relative_to: RelativeArg) -> 'WindowWrapperT':
        ...

    @abc.abstractmethod
    def relative_to_now(self) -> 'WindowWrapperT':
        ...

    @abc.abstractmethod
    def relative_to_field(self, field: _FieldDescriptorT) -> 'WindowWrapperT':
        ...

    @abc.abstractmethod
    def relative_to_stream(self) -> 'WindowWrapperT':
        ...

    @abc.abstractmethod
    def get_timestamp(self, event: EventT = None) -> float:
        ...

    @abc.abstractmethod
    def __getitem__(self, key: Any) -> WindowSetT:
        ...

    @abc.abstractmethod
    def keys(self) -> KeysView:
        ...

    @abc.abstractmethod
    def on_set_key(self, key: Any, value: Any) -> None:
        ...

    @abc.abstractmethod
    def on_del_key(self, key: Any) -> None:
        ...

    @abc.abstractmethod
    def as_ansitable(self, **kwargs: Any) -> str:
        ...

    @property
    def get_relative_timestamp(self) -> Optional[RelativeHandler]:
        ...

    @get_relative_timestamp.setter
    def get_relative_timestamp(self, relative_to: RelativeArg) -> None:
        ...



import pytest
from faust import web
from faust import Event, Stream, Table, Topic
from faust.assignor import PartitionAssignor
from faust.sensors import Sensor
from faust.transport.consumer import Consumer
from faust.transport.producer import Producer
from faust.types import Message, TP
from mode.utils.mocks import Mock

TP1 = TP('foo', 0)


@pytest.fixture
def message():
    return Mock(name='message', autospec=Message)


@pytest.fixture
def stream():
    return Mock(name='stream', autospec=Stream)


@pytest.fixture
def event():
    return Mock(name='event', autospec=Event)


@pytest.fixture
def topic():
    return Mock(name='topic', autospec=Topic)


@pytest.fixture
def table():
    return Mock(name='table', autospec=Table)


@pytest.fixture
def consumer():
    return Mock(name='consumer', autospec=Consumer)


@pytest.fixture
def producer():
    return Mock(name='producer', autospec=Producer)


@pytest.fixture
def assignor():
    return Mock(name='assignor', autospec=PartitionAssignor)


@pytest.fixture
def view():
    return Mock(name='view', autospec=web.View)


@pytest.fixture
def req():
    return Mock(name='request', autospec=web.Request)


@pytest.fixture
def response():
    return Mock(name='response', autospec=web.Response)


class test_Sensor:

    @pytest.fixture
    def sensor(self, *, app):
        return Sensor()

    def test_on_message_in(self, *, sensor, message):
        sensor.on_message_in(TP1, 3, message)

    def test_on_stream_event_in(self, *, sensor, stream, event):
        sensor.on_stream_event_in(TP1, 3, stream, event)

    def test_on_stream_event_out(self, *, sensor, stream, event):
        state = sensor.on_stream_event_in(TP1, 3, stream, event)
        sensor.on_stream_event_out(TP1, 3, stream, event, state)
        sensor.on_stream_event_out(TP1, 3, stream, event, None)

    def test_on_message_out(self, *, sensor, message):
        sensor.on_message_out(TP1, 3, message)

    def test_on_topic_buffer_full(self, *, sensor, topic):
        sensor.on_topic_buffer_full(topic)

    def test_on_table_get(self, *, sensor, table):
        sensor.on_table_get(table, 'key')

    def test_on_table_set(self, *, sensor, table):
        sensor.on_table_set(table, 'key', 'value')

    def test_on_table_del(self, *, sensor, table):
        sensor.on_table_del(table, 'key')

    def test_on_commit_initiated(self, *, sensor, consumer):
        sensor.on_commit_initiated(consumer)

    def test_on_commit_completed(self, *, sensor, consumer):
        sensor.on_commit_completed(consumer, Mock(name='state'))

    def test_on_send_initiated(self, *, sensor, producer):
        sensor.on_send_initiated(producer, 'topic', 'message', 30, 40)

    def test_on_send_completed(self, *, sensor, producer):
        sensor.on_send_completed(
            producer, Mock(name='state'), Mock(name='metadata'))

    def test_on_assignment(self, *, sensor, assignor):
        state = sensor.on_assignment_start(assignor)
        assert state['time_start']
        sensor.on_assignment_error(assignor, state, KeyError())
        sensor.on_assignment_completed(assignor, state)

    def test_on_rebalance(self, *, sensor, app):
        state = sensor.on_rebalance_start(app)
        assert state['time_start']
        sensor.on_rebalance_return(app, state)
        sensor.on_rebalance_end(app, state)

    def test_on_web_request(self, *, sensor, app, req, response, view):
        state = sensor.on_web_request_start(app, req, view=view)

        assert state['time_start']

        sensor.on_web_request_end(app, req, response, state, view=view)

    def test_on_send_error(self, *, sensor, producer):
        sensor.on_send_error(
            producer, KeyError('foo'), Mock(name='state'))

    def test_asdict(self, *, sensor):
        assert sensor.asdict() == {}


class test_SensorDelegate:

    @pytest.fixture
    def sensor(self):
        return Mock(name='sensor', autospec=Sensor)

    @pytest.fixture
    def sensors(self, *, app, sensor):
        sensors = app.sensors
        sensors.add(sensor)
        return sensors

    def test_remove(self, *, sensors, sensor):
        assert list(iter(sensors))
        sensors.remove(sensor)
        assert not list(iter(sensors))

    def test_on_message_in(self, *, sensors, sensor, message):
        sensors.on_message_in(TP1, 303, message)
        sensor.on_message_in.assert_called_once_with(TP1, 303, message)

    def test_on_stream_event_in_out(self, *, sensors, sensor, stream, event):
        state = sensors.on_stream_event_in(TP1, 303, stream, event)
        sensor.on_stream_event_in.assert_called_once_with(
            TP1, 303, stream, event)
        sensors.on_stream_event_out(TP1, 303, stream, event, state)
        sensor.on_stream_event_out.assert_called_once_with(
            TP1, 303, stream, event, state[sensor])

    def test_on_topic_buffer_full(self, *, sensors, sensor, topic):
        sensors.on_topic_buffer_full(topic)
        sensor.on_topic_buffer_full.assert_called_once_with(topic)

    def test_on_message_out(self, *, sensors, sensor, message):
        sensors.on_message_out(TP1, 303, message)
        sensor.on_message_out.assert_called_once_with(TP1, 303, message)

    def test_on_table_get(self, *, sensors, sensor, table):
        sensors.on_table_get(table, 'key')
        sensor.on_table_get.assert_called_once_with(table, 'key')

    def test_on_table_set(self, *, sensors, sensor, table):
        sensors.on_table_set(table, 'key', 'value')
        sensor.on_table_set.assert_called_once_with(table, 'key', 'value')

    def test_on_table_del(self, *, sensors, sensor, table):
        sensors.on_table_del(table, 'key')
        sensor.on_table_del.assert_called_once_with(table, 'key')

    def test_on_commit(self, *, sensors, sensor, consumer):
        state = sensors.on_commit_initiated(consumer)
        sensor.on_commit_initiated.assert_called_once_with(consumer)

        sensors.on_commit_completed(consumer, state)
        sensor.on_commit_completed.assert_called_once_with(
            consumer, state[sensor])

    def test_on_send(self, *, sensors, sensor, producer):
        metadata = Mock(name='metadata')
        state = sensors.on_send_initiated(
            producer, 'topic', 'message', 303, 606)
        sensor.on_send_initiated.assert_called_once_with(
            producer, 'topic', 'message', 303, 606)

        sensors.on_send_completed(producer, state, metadata)
        sensor.on_send_completed.assert_called_once_with(
            producer, state[sensor], metadata)

        exc = KeyError('foo')
        sensors.on_send_error(producer, exc, state)
        sensor.on_send_error.assert_called_once_with(
            producer, exc, state[sensor])

    def test_on_assignment(self, *, sensors, sensor, assignor):
        state = sensors.on_assignment_start(assignor)
        sensor.on_assignment_start.assert_called_once_with(assignor)

        sensors.on_assignment_completed(assignor, state)
        sensor.on_assignment_completed.assert_called_once_with(
            assignor, state[sensor])

        exc = KeyError('bar')
        sensors.on_assignment_error(assignor, state, exc)
        sensor.on_assignment_error.assert_called_once_with(
            assignor, state[sensor], exc)

    def test_on_rebalance(self, *, sensors, sensor, app):
        state = sensors.on_rebalance_start(app)
        sensor.on_rebalance_start.assert_called_once_with(app)

        sensors.on_rebalance_return(app, state)
        sensor.on_rebalance_return.assert_called_once_with(
            app, state[sensor])

        sensors.on_rebalance_end(app, state)
        sensor.on_rebalance_end.assert_called_once_with(
            app, state[sensor])

    def test_on_web_request(self, *,
                            sensors,
                            sensor,
                            app,
                            req,
                            response,
                            view):
        state = sensors.on_web_request_start(app, req, view=view)
        sensor.on_web_request_start.assert_called_once_with(
            app, req, view=view)

        sensors.on_web_request_end(app, req, response, state, view=view)
        sensor.on_web_request_end.assert_called_once_with(
            app, req, response, state[sensor], view=view)

    def test_repr(self, *, sensors):
        assert repr(sensors)



#!/usr/bin/env python

from random import random
from datetime import datetime
import faust

app = faust.App('tz_unaware', broker='kafka://localhost:9092')


class Model(faust.Record, serializer='json'):
    random: float


tz_unaware_topic = app.topic('tz_unaware_topic', value_type=Model)


@app.agent(tz_unaware_topic)
async def consume(stream):
    async for record in stream:
        print(f'record: {record}')


@app.crontab('*/1 * * * *', on_leader=True)
async def publish_every_minute():
    print('-- We should send a message every minute --')
    print(f'Sending message at: {datetime.now()}')
    msg = Model(random=round(random(), 2))
    await tz_unaware_topic.send(value=msg)


if __name__ == '__main__':
    app.main()



import pytest
from datetime import datetime, timezone
from faust.utils._iso8601_python import InvalidTZ, parse, parse_tz


def test_python():
    dt1 = datetime.now().astimezone(timezone.utc)
    dt2 = parse(dt1.isoformat())
    assert dt1 == dt2


def test_timezone_no_sep():
    dt = parse('2018-12-04T19:36:08-0500')
    assert dt.tzinfo
    assert str(dt.tzinfo) == 'UTC-05:00'


def test_parse_error():
    with pytest.raises(ValueError):
        parse('foo')


@pytest.mark.parametrize('tz', [
    'Z',
    '+00:10',
    '-01:20',
    '+0300',
    '-0600',
])
def test_parse_tz(tz):
    assert parse_tz(tz) is not None


def test_parse_tz__no_match():
    with pytest.raises(InvalidTZ):
        parse_tz('foo')



"""Agents."""
from .agent import Agent, AgentFun, AgentT, SinkT, current_agent
from .manager import AgentManager, AgentManagerT
from .replies import ReplyConsumer

__all__ = [
    'Agent',
    'AgentFun',
    'AgentT',
    'SinkT',
    'AgentManager',
    'AgentManagerT',
    'ReplyConsumer',
    'current_agent',
]



import asyncio
import logging
import warnings
from pathlib import Path

import pytest
from faust import Sensor
from faust.worker import Worker
from faust.utils import terminal
from mode.utils.logging import CompositeLogger
from mode.utils.trees import Node
from mode.utils.mocks import AsyncMock, Mock, patch
from yarl import URL


class CoroEq:

    def __init__(self, coro):
        self.coro = coro

    def __eq__(self, other):
        return other.cr_code.co_name == self.coro.__name__


class test_Worker:

    @pytest.fixture
    def worker(self, app):
        return Worker(app)

    def test_constructor(self, app):
        w = Worker(app)
        assert w.app is app
        assert w.sensors == set()
        assert w.workdir == Path.cwd()
        assert isinstance(w.spinner, terminal.Spinner)
        w2 = Worker(app, redirect_stdouts=False)
        assert not w2.redirect_stdouts
        w3 = Worker(app, redirect_stdouts_level='DEBUG')
        assert w3.redirect_stdouts_level == 10
        w4 = Worker(app, logging_config={'foo': 1})
        assert w4.logging_config == {'foo': 1}

    def test_set_sensors(self, app):
        assert Worker(app, sensors=[1, 2]).sensors == {1, 2}

    def test_set_workdir(self, app):
        assert Worker(app, workdir='/foo').workdir == Path('/foo')

    @pytest.mark.asyncio
    async def test_on_start(self, worker):
        await worker.on_start()

    @pytest.mark.asyncio
    async def test_on_siginit(self, worker):
        with warnings.catch_warnings():
            with patch('asyncio.ensure_future') as ensure_future:
                worker._on_sigint()
                assert worker._shutdown_immediately
                assert worker.spinner.stopped
                ensure_future.assert_called_with(
                    CoroEq(worker._stop_on_signal), loop=worker.loop)
                coro = ensure_future.call_args[0][0]
        asyncio.ensure_future(coro).cancel()  # silence warning

    @pytest.mark.asyncio
    async def test_on_siginit__no_spinner(self, worker, loop):
        worker.spinner = None
        with patch('asyncio.ensure_future') as ensure_future:
            worker._on_sigint()
            coro = ensure_future.call_args[0][0]
        asyncio.ensure_future(coro, loop=loop).cancel()

    @pytest.mark.asyncio
    async def test__on_sigterm(self, worker):
        with patch('asyncio.ensure_future') as ensure_future:
            worker._on_sigterm()
            assert worker._shutdown_immediately
            assert worker.spinner.stopped
            ensure_future.assert_called_with(
                CoroEq(worker._stop_on_signal), loop=worker.loop)
            coro = ensure_future.call_args[0][0]
        asyncio.ensure_future(coro).cancel()  # silence warning

    @pytest.mark.asyncio
    async def test_on_startup_finished__shutdown_requested(self, worker):
        worker._shutdown_immediately = True
        worker._on_shutdown_immediately = Mock(name='on_shutdown_immediately')
        await worker.on_startup_finished()
        worker._on_shutdown_immediately.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_startup_finished(self, worker):
        worker.maybe_start_blockdetection = AsyncMock(name='maybe_start_block')
        worker._on_startup_end_spinner = Mock(name='on_startup_end_spinner')
        await worker.on_startup_finished()
        worker.maybe_start_blockdetection.assert_called_once_with()
        worker._on_startup_end_spinner.assert_called_once_with()

    def test_on_startup_end_spinner(self, worker):
        spinner = worker.spinner = Mock(
            name='spinner',
            autospec=terminal.Spinner,
        )
        spinner.file.isatty.return_value = True
        worker.say = Mock(name='say')
        worker._on_startup_end_spinner()
        spinner.finish.assert_called_once_with()
        worker.say.assert_called_once_with(' 😊')

    def test_on_startup_end_spinner__no_spinner(self, worker):
        worker.spinner = None
        worker.log = Mock(name='log', spec=CompositeLogger)
        worker._on_startup_end_spinner()
        worker.log.info.assert_called_once_with('Ready')

    def test_on_startup_end_spinner__notatty(self, worker):
        spinner = worker.spinner = Mock(
            name='spinner',
            autospec=terminal.Spinner,
        )
        spinner.file.isatty.return_value = False
        worker.say = Mock(name='say')
        worker._on_startup_end_spinner()
        spinner.finish.assert_called_once_with()
        worker.say.assert_called_once_with(' OK ^')

    def test_on_shutdown_immediately(self, worker):
        worker.say = Mock(name='say')
        worker._on_shutdown_immediately()
        worker.say.assert_called_once_with('')

    def test_on_init_dependencies(self, worker, app):
        app.beacon = Mock(name='app.beacon', autospec=Node)
        deps = worker.on_init_dependencies()
        assert list(deps) == list(worker.services) + [app]
        app.beacon.reattach.assert_called_once_with(worker.beacon)
        assert app.on_startup_finished == worker.on_startup_finished

    def test_on_init_dependencies__sensors_to_app(self, worker, app):
        s1 = Mock(name='S1', autospec=Sensor)
        s2 = Mock(name='S2', autospec=Sensor)
        worker.sensors = {s1, s2}
        worker.on_init_dependencies()
        assert app.sensors._sensors.issubset(worker.sensors)

    @pytest.mark.asyncio
    async def test_on_first_start(self, worker):
        worker.change_workdir = Mock(name='change_workdir')
        worker.autodiscover = Mock(name='autodiscover')
        worker.default_on_first_start = AsyncMock(name='on_first_start')
        await worker.on_first_start()
        worker.change_workdir.assert_called_once_with(worker.workdir)
        worker.autodiscover.assert_called_once_with()
        worker.default_on_first_start.assert_called_once_with()

    def test_change_workdir(self, worker):
        with patch('os.chdir') as chdir:
            p = Path('baz')
            worker.change_workdir(p)
            chdir.assert_called_once_with(p.absolute())

    def test_change_workdir__already_cwd(self, worker):
        with patch('os.chdir') as chdir:
            p = Path.cwd()
            worker.change_workdir(p)
            chdir.assert_not_called()

    def test_autodiscover(self, worker):
        worker.app.conf.autodiscover = True
        worker.app.discover = Mock(name='discover')
        worker.autodiscover()
        worker.app.discover.assert_called_once_with()

    def test_autodiscover__disabled(self, worker):
        worker.app.conf.autodiscover = False
        worker.app.discover = Mock(name='discover')
        worker.autodiscover()
        worker.app.discover.assert_not_called()

    def test_setproctitle(self, worker, app):
        with patch('faust.worker.setproctitle') as setproctitle:
            worker._setproctitle('foo')
            setproctitle.assert_called_with(
                f'[Faust:Worker] -foo- testid -p {app.conf.web_port} '
                f'{app.conf.datadir.absolute()}')

    def test_proc_ident(self, worker, app):
        assert (worker._proc_ident() ==
                f'testid -p {app.conf.web_port} {app.conf.datadir.absolute()}')

    def test_proc_web_ident__unix(self, worker, app):
        worker.app.conf.web_transport = URL('unix:')
        assert worker._proc_web_ident() == str(URL('unix:'))

    def test_proc_web_ident__tcp(self, worker):
        worker.app.conf.web_transport_scheme = 'tcp'
        assert worker._proc_web_ident() == '-p 6066'

    def test_on_worker_shutdown(self, worker):
        worker.spinner = None
        worker._say = Mock(name='say')
        worker.on_worker_shutdown()
        worker.spinner = Mock(name='spinner')
        worker.on_worker_shutdown()
        worker.spinner.reset.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_execute(self, worker):
        worker._setproctitle = Mock(name='setproctitle')
        worker.spinner = Mock(
            name='spinner',
            autospec=terminal.Spinner,
        )
        worker._say = Mock(name='say')
        await worker.on_execute()
        worker._setproctitle.assert_called_with('init')
        worker._say.assert_called_with('starting➢ ', end='', flush=True)
        worker.spinner = None
        await worker.on_execute()

    def test_on_setup_root_logger(self, worker):
        worker._disable_spinner_if_level_below_WARN = Mock(name='dd')
        worker._setup_spinner_handler = Mock(name='ss')
        logger = Mock(name='logger', autospec=logging.Logger)
        worker.on_setup_root_logger(logger, logging.INFO)
        worker._disable_spinner_if_level_below_WARN.assert_called_with(
            logging.INFO)
        worker._setup_spinner_handler.assert_called_with(logger, logging.INFO)

    @pytest.mark.parametrize('loglevel,expected', [
        (None, True),
        (logging.CRITICAL, True),
        (logging.ERROR, True),
        (logging.WARN, True),
        (logging.INFO, False),
        (logging.DEBUG, False),
    ])
    def test_disable_spinner_if_level_below_WARN(self, loglevel, expected,
                                                 worker):
        worker._disable_spinner_if_level_below_WARN(loglevel)
        if expected:
            assert worker.spinner
        else:
            assert worker.spinner is None

    def test_setup_spinner_handler(self, worker):
        logger = Mock(name='logger', autospec=logging.Logger)
        logger.handlers = [Mock(name='handler', autospec=logging.Handler)]
        with patch('faust.utils.terminal.SpinnerHandler') as SpinnerHandler:
            worker._setup_spinner_handler(logger, logging.INFO)
            logger.handlers[0].setLevel.assert_called_with(logging.INFO)
            SpinnerHandler.assert_called_once_with(
                worker.spinner, level=logging.DEBUG)
            logger.addHandler.assert_called_once_with(SpinnerHandler())
            logger.setLevel.assert_called_once_with(logging.DEBUG)

    def test_setup_spinner_handler__when_no_spinner(self, worker):
        worker.spinner = None
        worker._setup_spinner_handler(
            Mock(
                name='logger',
                autospec=logging.Logger,
            ),
            logging.INFO,
        )



from typing import Any, Callable, Iterator, List
from mode.utils.objects import cached_property
import faust
from faust.types import RecordMetadata
from . import config
from . import producer
from .reports.checks import Check, SystemChecks
from .reports.logging import LogHandler, LogPusher

__all__ = ['ProducerFun', 'StressApp', 'create_stress_app']

ProducerFun = Callable[[int], Iterator[RecordMetadata]]


class StressApp(faust.App):
    stress_producers: List[ProducerFun]
    count_received_events: int

    #: Status page reports this number.
    #: and report_progress background thread updates it as tests fail.
    faults: int = 0

    unassigned: bool = False

    def __init__(self, *args, **kwargs):
        loghandler = LogHandler(self)
        kwargs['loghandlers'] = [loghandler]
        super().__init__(*args, **kwargs)
        self.stress_producers = []
        self.count_received_events = 0

    async def on_start(self) -> None:
        from .reports import checks
        self.system_checks.add(
            checks.Increasing(
                'events_total',
                get_value=lambda: self.monitor.events_total,
            ),
        )
        await self.add_runtime_dependency(self.system_checks)
        await self.add_runtime_dependency(self.logpusher)

    def add_system_check(self, check: Check) -> None:
        self.system_checks.add(check)

    def register_stress_producer(self, fun: ProducerFun):
        self.stress_producers.append(fun)
        return fun

    @cached_property
    def system_checks(self) -> SystemChecks:
        return SystemChecks(self)

    @cached_property
    def logpusher(self) -> LogPusher:
        return LogPusher(self)


def create_app(name, origin, base=faust.App, **kwargs: Any) -> faust.App:
    return base(
        name,
        origin=origin,
        broker=config.broker,
        store=config.store,
        topic_partitions=config.topic_partitions,
        loghandlers=config.loghandlers(),
        autodiscover=True,
        **kwargs)


def create_stress_app(name, origin, **kwargs: Any) -> StressApp:
    app = create_app(name, origin, base=StressApp, **kwargs)
    producer.install_produce_command(app)
    return app



import pytest
from mode.utils.mocks import AsyncMock, Mock, call
from faust.transport.producer import Producer, ProducerBuffer


class test_ProducerBuffer:

    @pytest.fixture()
    def buf(self):
        return ProducerBuffer()

    def test_put(self, *, buf):
        fut = Mock(name='future_message')
        buf.pending = Mock(name='pending')
        buf.put(fut)

        buf.pending.put_nowait.assert_called_once_with(fut)

    @pytest.mark.asyncio
    async def test_on_stop(self, *, buf):
        buf.flush = AsyncMock(name='flush')
        await buf.on_stop()
        buf.flush.coro.assert_called_once_with()

    @pytest.mark.asyncio
    async def test__send_pending(self, *, buf):
        fut = Mock(name='future_message')
        fut.message.channel.publish_message = AsyncMock()
        await buf._send_pending(fut)
        fut.message.channel.publish_message.coro.assert_called_once_with(
            fut, wait=False,
        )

    @pytest.mark.asyncio
    async def test__handle_pending(self, *, buf):
        buf.pending = Mock(get=AsyncMock())
        buf._send_pending = AsyncMock()

        async def on_send(fut):
            if buf._send_pending.call_count >= 3:
                buf._stopped.set()

        buf._send_pending.side_effect = on_send

        await buf._handle_pending(buf)

        buf._send_pending.assert_has_calls([
            call(buf.pending.get.coro.return_value),
            call(buf.pending.get.coro.return_value),
        ])

    @pytest.mark.asyncio
    async def test_wait_until_ebb(self, *, buf):
        buf.max_messages = 10
        buf._send_pending = AsyncMock()
        self._put(buf, range(20))
        assert buf.size == 20

        await buf.wait_until_ebb()
        assert list(buf.pending._queue) == list(range(10, 20))
        assert buf.size == 10

        await buf.wait_until_ebb()
        assert list(buf.pending._queue) == list(range(10, 20))
        assert buf.size == 10

    @pytest.mark.asyncio
    async def test_flush(self, *, buf):
        buf._send_pending = AsyncMock()
        assert not buf.size
        await buf.flush()

        self._put(buf, range(10))
        assert buf.size == 10

        await buf.flush()
        assert not buf.size

    def _put(self, buf, items):
        for item in items:
            buf.pending.put_nowait(item)

    @pytest.mark.asyncio
    async def test_flush_atmost(self, *, buf):
        buf._send_pending = AsyncMock()
        assert await buf.flush_atmost(10) == 0

        self._put(buf, range(3))
        assert buf.size == 3
        assert await buf.flush_atmost(10) == 3

        self._put(buf, range(10))
        assert buf.size == 10
        assert (await buf.flush_atmost(4)) == 4
        assert buf.size == 6

        assert (await buf.flush_atmost(6)) == 6
        assert not buf.size


class ProducerTests:

    @pytest.mark.asyncio
    async def test_send(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.send('topic', 'key', 'value', 1, None, {})

    def test_send_soon(self, *, producer):
        producer.buffer = Mock(name='buffer')
        fut = Mock(name='fut')
        producer.send_soon(fut)
        producer.buffer.put.assert_called_once_with(fut)

    @pytest.mark.asyncio
    async def test_flush(self, *, producer):
        await producer.flush()

    @pytest.mark.asyncio
    async def test_send_and_wait(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.send_and_wait('topic', 'key', 'value', 1, None, {})

    @pytest.mark.asyncio
    async def test_create_topic(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.create_topic('topic', 1, 1)

    def test_key_partition(self, *, producer):
        with pytest.raises(NotImplementedError):
            producer.key_partition('topic', 'key')

    @pytest.mark.asyncio
    async def test_begin_transaction(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.begin_transaction('tid')

    @pytest.mark.asyncio
    async def test_commit_transaction(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.commit_transaction('tid')

    @pytest.mark.asyncio
    async def test_abort_transaction(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.abort_transaction('tid')

    @pytest.mark.asyncio
    async def test_stop_transaction(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.stop_transaction('tid')

    @pytest.mark.asyncio
    async def test_maybe_begin_transaction(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.maybe_begin_transaction('tid')

    @pytest.mark.asyncio
    async def test_commit_transactions(self, *, producer):
        with pytest.raises(NotImplementedError):
            await producer.commit_transactions({}, 'gid')

    def test_supports_headers(self, *, producer):
        assert not producer.supports_headers()


class test_Producer(ProducerTests):

    @pytest.fixture
    def producer(self, *, app):
        return Producer(app.transport)



"""Using :pypi:`terminaltables` to draw ANSI tables."""
import sys

from operator import itemgetter
from typing import (
    Any,
    Callable,
    IO,
    Iterable,
    List,
    Mapping,
    Sequence,
    Type,
    cast,
)

from mode.utils import logging
from mode.utils import text
from mode.utils.compat import isatty

from terminaltables import AsciiTable, SingleTable
from terminaltables.base_table import BaseTable as Table

__all__ = ['Table', 'TableDataT', 'table', 'logtable']

TableDataT = Sequence[Sequence[str]]


def table(data: TableDataT,
          *,
          title: str,
          target: IO = None,
          tty: bool = None,
          **kwargs: Any) -> Table:
    """Create suitable :pypi:`terminaltables` table for target.

    Arguments:
        data (Sequence[Sequence[str]]): Table data.

        target (IO): Target should be the destination output file
                     for your table, and defaults to :data:`sys.stdout`.
                     ANSI codes will be used if the target has a controlling
                     terminal, but not otherwise, which is why it's important
                     to pass the correct output file.
    """
    if target is None:
        target = sys.stdout
    if tty is None:
        tty = isatty(target)
    if tty is None:
        tty = False
    return _get_best_table_type(tty)(data, title=title, **kwargs)


def logtable(data: TableDataT,
             *,
             title: str,
             target: IO = None,
             tty: bool = None,
             headers: Sequence[str] = None,
             **kwargs: Any) -> str:
    """Prepare table for logging.

    Will use ANSI escape codes if the log file is a tty.
    """
    if tty is None:
        tty = logging.LOG_ISATTY
    if headers:
        data = [headers] + list(data)
    return table(data, title=title, target=target, tty=tty, **kwargs).table


def _get_best_table_type(tty: bool) -> Type[Table]:
    return SingleTable if tty else AsciiTable


def dict_as_ansitable(d: Mapping,
                      *,
                      key: str = 'Key',
                      value: str = 'Value',
                      sort: bool = False,
                      sortkey: Callable[[Any], Any] = itemgetter(0),
                      target: IO = sys.stdout,
                      title: str = None) -> str:
    header = [text.title(key), text.title(value)]
    data = cast(Iterable[List[str]], d.items())
    data = sorted(data, key=sortkey) if sort else list(data)
    if sort:
        data = sorted(data, key=sortkey)
    return table(
        [header] + list(data),
        title=text.title(title) if title is not None else '',
        target=target,
    ).table



import abc
import sys
from collections import UserString
from contextlib import contextmanager
from decimal import Decimal
from types import FrameType
from typing import (
    Any,
    Generic,
    Iterator,
    Optional,
    Type,
    TypeVar,
    cast,
    no_type_check,
)
from mode.utils.locals import LocalStack
from faust.exceptions import SecurityError

T = TypeVar('T')

_getframe = getattr(sys, 'emarfteg_'[::-1])

AllowedStack: LocalStack[bool] = LocalStack()


@contextmanager
def allow_protected_vars() -> Iterator:
    with AllowedStack.push(True):
        yield


class Tag(Generic[T]):
    value: T
    field: Optional[str]

    is_secret: bool = False
    is_sensitive: bool = False
    is_personal: bool = False

    def __init__(self, value: T, *,
                 field: str = None):
        if isinstance(value, Tag):
            raise SecurityError('Cannot wrap: value is already tagged')
        self._value = value
        self.field = field

    def __set_name__(self, owner: Type, name: str) -> None:
        self.field = name

    def get_value(self) -> T:
        return self._value

    def __repr__(self) -> str:
        return f'<{self._name}: {self.field}@{id(self):#x}>'

    @abc.abstractmethod
    def __str__(self) -> str:
        ...

    @abc.abstractmethod
    def __format__(self, format_spec: str) -> str:
        ...

    @property
    def _name(self) -> str:
        return type(self).__name__


class OpaqueTag(Tag[T]):

    def __str__(self) -> str:
        raise SecurityError(f'Attempt to use {self._name} data as a string')

    def __format__(self, format_spec: str) -> str:
        raise SecurityError(f'Attempt to use {self._name} data as a string')


class TransparentTag(Tag[T]):

    def __str__(self) -> str:
        return str(self._prepare_value())

    def _prepare_value(self) -> T:
        return self._value

    def __format__(self, format_spec: str) -> str:
        return self._prepare_value().__format__(format_spec)


class _FrameLocal(UserString, Generic[T]):

    _field_name: str
    _tag_type: str
    _frame: str
    _value: T

    def __init__(self, value: T, *,
                 field_name: str = '<unknown field>',
                 tag_type: str = '<unknown tag>') -> None:
        self._value = value
        self._frame = self._frame_ident(_getframe().f_back.f_back)
        self._field_name = field_name
        self._tag_type = tag_type

    def _access_value(self) -> T:
        if AllowedStack.top:
            return self._value
        current_frame = self._frame_ident(
            _getframe().f_back.f_back.f_back)
        import traceback
        traceback.print_stack()
        if current_frame == self._frame:
            return self._value
        else:
            raise SecurityError(
                f'Protected {self._tag_type} value from '
                f'field {self._field_name} accessed outside origin frame.')

    def __repr__(self) -> str:
        val = self._value
        return f'<Protected {type(val)}: {val!r}@{id(self):#x}>'

    def _frame_ident(self, frame: FrameType) -> str:
        return '::'.join(map(str, [
            # cannot rely on id alone as memory addresses are reused
            id(frame),
            frame.f_code.co_filename,
            frame.f_code.co_name,
        ]))

    @property
    def data(self) -> T:  # type: ignore
        return self._access_value()


class Personal(OpaqueTag[T]):
    is_personal = True
    FrameLocal: Type[_FrameLocal] = _FrameLocal

    def get_value(self) -> T:
        if AllowedStack.top:
            return self._value
        else:
            return cast(T, self.FrameLocal(
                self._value,
                field_name=self.field or '<unknown field>',
                tag_type=self._name.lower(),
            ))

    @no_type_check
    def __class_getitem__(self, params: Any) -> Any:
        if not issubclass(params, (str, bytes)):
            raise TypeError(f'Personal only supports str/bytes not {params!r}')
        return super().__class_getitem__(params)


class Secret(TransparentTag[T]):
    is_secret = True

    mask: str = '***********'

    def _prepare_value(self) -> T:
        return cast(T, self.mask)


class Sensitive(OpaqueTag[T]):
    is_sensitive = True

    FrameLocal: Type[_FrameLocal] = _FrameLocal

    def get_value(self) -> T:
        return cast(T, self.FrameLocal(
            self._value,
            field_name=self.field or '<unknown field>',
            tag_type=self._name.lower(),
        ))

    def __class_getitem__(self, params: Any) -> Any:
        if not issubclass(params, (str, bytes)):
            raise TypeError(f'Personal only supports str/bytes not {params!r}')

        # bypass mypy bug by using getattr
        sup = getattr(super(), '__class_getitem__')  # noqa: B009
        return sup(params)


class _PIIstr(str):
    ...


class _PIIbytes(bytes):
    ...


class _PIIint(int):
    ...


class _PIIfloat(float):
    ...


class _PIIDecimal(Decimal):
    ...



import logging as _logging
import os
import pytest
import faust
from copy import copy
from typing import Dict, IO, NamedTuple, Union
from faust.web.cache.backends.memory import CacheStorage
from faust.utils.tracing import set_current_span
from mode.utils.logging import setup_logging
from mode.utils.mocks import AsyncMock, Mock


class AppMarks(NamedTuple):
    name: str = 'funtest'
    store: str = 'memory://'
    cache: str = 'memory://'


def create_appmarks(name='funtest',
                    store='memory://',
                    cache='memory://',
                    **rest):
    options = AppMarks(
        name=name,
        store=store,
        cache=cache,
    )
    return options, rest


@pytest.yield_fixture()
def app(event_loop, request):
    os.environ.pop('F_DATADIR', None)
    os.environ.pop('FAUST_DATADIR', None)
    os.environ.pop('F_WORKDIR', None)
    os.environ.pop('FAUST_WORKDIR', None)
    marks = request.node.get_closest_marker('app')
    options, rest = create_appmarks(
        **((marks.kwargs or {}) if marks else {}))
    app = faust.App(
        options.name,
        store=options.store,
        cache=options.cache,
        **rest,
    )
    app.finalize()
    set_current_span(None)
    try:
        yield app
    finally:
        assert app.tracer is None


@pytest.fixture()
def web(app):
    app.web.init_server()
    return app.web


class LoggingMarks(NamedTuple):
    logfile: Union[str, IO] = None
    loglevel: Union[str, int] = 'info'
    logging_config: Dict = None


@pytest.yield_fixture()
def logging(request):
    marks = request.node.get_closest_marker('logging')
    options = LoggingMarks(**{
        **{'logfile': None,
           'loglevel': 'info',
           'logging_config': None},
        **((marks.kwargs or {}) if marks else {}),
    })
    _logging._acquireLock()
    try:
        prev_state = copy(_logging.Logger.manager.loggerDict)
        prev_handlers = copy(_logging.root.handlers)
    finally:
        _logging._releaseLock()
    try:
        setup_logging(
            logfile=options.logfile,
            loglevel=options.loglevel,
            logging_config=options.logging_config,
        )
        yield
    finally:
        _logging._acquireLock()
        try:
            _logging.Logger.manager.loggerDict = prev_state
            _logging.root.handlers = prev_handlers
        finally:
            _logging._releaseLock()


@pytest.fixture()
def mocked_redis(*, event_loop, monkeypatch):
    import aredis

    storage = CacheStorage()

    client_cls = Mock(
        name='StrictRedis',
        return_value=Mock(
            autospec=aredis.StrictRedis,
            ping=AsyncMock(),
            get=AsyncMock(side_effect=storage.get),
            set=AsyncMock(side_effect=storage.set),
            setex=AsyncMock(side_effect=storage.setex),
            delete=AsyncMock(side_effect=storage.delete),
            ttl=AsyncMock(side_effect=storage.ttl),
        ),
    )
    client_cls.storage = storage
    monkeypatch.setattr('aredis.StrictRedis', client_cls)
    return client_cls


def pytest_configure(config):
    config.addinivalue_line(
        'markers', 'app: App instance to use for tests',
    )
    config.addinivalue_line(
        'markers', 'logging: Configure logging setup to use for tests',
    )



"""Program ``faust`` (umbrella command)."""

# Note: The command options above are defined in .cli.base.builtin_options
from .agents import agents
from .base import call_command, cli
from .clean_versions import clean_versions
from .completion import completion
from .livecheck import livecheck
from .model import model
from .models import models
from .reset import reset
from .send import send
from .tables import tables
from .worker import worker

__all__ = [
    'agents',
    'call_command',
    'clean_versions',
    'cli',
    'completion',
    'livecheck',
    'model',
    'models',
    'reset',
    'send',
    'tables',
    'worker',
]



import asyncio
import pytest
from faust import App
from faust.app._attached import Attachments
from faust.exceptions import AlreadyConfiguredWarning
from faust.tables.manager import TableManager
from faust.transport.base import Producer, Transport
from intervaltree import IntervalTree
from faust.transport.consumer import (
    Consumer,
    ConsumerThread,
    Fetcher,
    ProducerSendError,
    ThreadDelegateConsumer,
    TransactionManager,
)
from faust.transport.conductor import Conductor
from faust.types import Message, TP
from mode import Service
from mode.threads import MethodQueue
from mode.utils.futures import done_future
from mode.utils.mocks import ANY, AsyncMock, Mock, call, patch
from sortedcontainers import SortedSet

TP1 = TP('foo', 0)
TP2 = TP('foo', 1)
TP3 = TP('bar', 3)


class test_Fetcher:

    @pytest.fixture
    def consumer(self):
        return Mock(
            name='consumer',
            autospec=Consumer,
            _drain_messages=AsyncMock(),
        )

    @pytest.fixture
    def fetcher(self, *, app, consumer):
        fetcher = Fetcher(app)
        app.consumer = consumer
        # some weird pytest-asyncio thing
        fetcher.loop = asyncio.get_event_loop()
        return fetcher

    @pytest.mark.asyncio
    async def test_fetcher(self, *, fetcher, app):
        await fetcher._fetcher(fetcher)
        app.consumer._drain_messages.assert_called_once_with(fetcher)

    @pytest.mark.asyncio
    async def test_fetcher__raises_CancelledError(self, *, fetcher, app):
        app.consumer._drain_messages.side_effect = asyncio.CancelledError
        await fetcher._fetcher(fetcher)
        app.consumer._drain_messages.assert_called_once_with(fetcher)

    @pytest.mark.asyncio
    async def test_on_stop__no_drainer(self, *, fetcher):
        fetcher._drainer = None
        await fetcher.on_stop()

    @pytest.mark.asyncio
    async def test_on_stop__drainer_done(self, *, fetcher):
        fetcher._drainer = Mock(done=Mock(return_value=True))
        await fetcher.on_stop()

    @pytest.mark.asyncio
    async def test_on_stop_drainer__drainer_done2(self, *, fetcher):
        fetcher._drainer = Mock(done=Mock(return_value=False))
        with patch('asyncio.wait_for', AsyncMock()) as wait_for:
            wait_for.coro.return_value = None
            await fetcher.on_stop()
        fetcher._drainer.cancel.assert_called_once_with()
        assert wait_for.call_count

    @pytest.mark.asyncio
    async def test_on_stop__drainer_pending(self, *, fetcher):
        fetcher._drainer = Mock(done=Mock(return_value=False))
        with patch('asyncio.wait_for', AsyncMock()) as wait_for:
            await fetcher.on_stop()
            wait_for.assert_called_once_with(
                fetcher._drainer, timeout=1.0,
            )

    @pytest.mark.asyncio
    async def test_on_stop__drainer_raises_StopIteration(self, *, fetcher):
        fetcher._drainer = Mock(done=Mock(return_value=False))
        with patch('asyncio.wait_for', AsyncMock()) as wait_for:
            wait_for.side_effect = StopIteration()
            await fetcher.on_stop()
            wait_for.assert_called_once_with(
                fetcher._drainer, timeout=1.0,
            )

    @pytest.mark.asyncio
    async def test_on_stop__drainer_raises_CancelledError(self, *, fetcher):
        fetcher._drainer = Mock(done=Mock(return_value=False))
        with patch('asyncio.wait_for', AsyncMock()) as wait_for:
            wait_for.coro.side_effect = asyncio.CancelledError()
            await fetcher.on_stop()
            wait_for.assert_called_once_with(
                fetcher._drainer, timeout=1.0,
            )

    @pytest.mark.asyncio
    async def test_on_stop__drainer_raises_TimeoutError(self, *, fetcher):
        fetcher._drainer = Mock(done=Mock(return_value=False))
        with patch('asyncio.wait_for', AsyncMock()) as wait_for:
            wait_for.coro.side_effect = [
                asyncio.TimeoutError(),
                asyncio.TimeoutError(),
                None,
            ]
            await fetcher.on_stop()
            wait_for.assert_called_with(
                fetcher._drainer, timeout=1.0,
            )
            assert wait_for.call_count == 3


class test_TransactionManager:

    @pytest.fixture()
    def consumer(self):
        return Mock(
            name='consumer',
            spec=Consumer,
        )

    @pytest.fixture()
    def producer(self):
        return Mock(
            name='producer',
            spec=Producer,
            create_topic=AsyncMock(),
            stop_transaction=AsyncMock(),
            maybe_begin_transaction=AsyncMock(),
            commit_transactions=AsyncMock(),
            send=AsyncMock(),
            flush=AsyncMock(),
        )

    @pytest.fixture()
    def transport(self, *, app):
        return Mock(
            name='transport',
            spec=Transport,
            app=app,
        )

    @pytest.fixture()
    def manager(self, *, consumer, producer, transport):
        return TransactionManager(
            transport,
            consumer=consumer,
            producer=producer,
        )

    @pytest.mark.asyncio
    async def test_on_partitions_revoked(self, *, manager):
        manager.flush = AsyncMock()
        await manager.on_partitions_revoked({TP1})
        manager.flush.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, manager):
        TP3_group = 0
        TP2_group = 2
        manager.app.assignor._topic_groups = {
            TP3.topic: TP3_group,
            TP2.topic: TP2_group,
        }

        assert TP3.topic != TP2.topic  # must be different topics
        manager._stop_transactions = AsyncMock()
        manager._start_transactions = AsyncMock()

        assigned = {TP2}
        revoked = {TP3}
        newly_assigned = {TP2}
        await manager.on_rebalance(assigned, revoked, newly_assigned)

        manager._stop_transactions.assert_called_once_with(
            [f'{manager.app.conf.id}-{TP3_group}-{TP3.partition}'])
        manager._start_transactions.assert_called_once_with(
            [f'{manager.app.conf.id}-{TP2_group}-{TP2.partition}'])

        await manager.on_rebalance(set(), set(), set())

    @pytest.mark.asyncio
    async def test__stop_transactions(self, *, manager, producer):
        await manager._stop_transactions(['0-0', '1-0'])
        producer.stop_transaction.assert_has_calls([
            call('0-0'),
            call.coro('0-0'),
            call('1-0'),
            call.coro('1-0'),
        ])

    @pytest.mark.asyncio
    async def test_start_transactions(self, *, manager, producer):
        manager._start_new_producer = AsyncMock()
        await manager._start_transactions(['0-0', '1-0'])
        producer.maybe_begin_transaction.assert_has_calls([
            call('0-0'),
            call.coro('0-0'),
            call('1-0'),
            call.coro('1-0'),
        ])

    @pytest.mark.asyncio
    async def test_send(self, *, manager, producer):
        manager.app.assignor._topic_groups = {
            't': 3,
        }
        manager.consumer.key_partition.return_value = 1

        await manager.send(
            't', 'k', 'v', partition=None, headers=None, timestamp=None)
        manager.consumer.key_partition.assert_called_once_with('t', 'k', None)
        producer.send.assert_called_once_with(
            't', 'k', 'v', 1, None, None, transactional_id='testid-3-1',
        )

    @pytest.mark.asyncio
    async def test_send__topic_not_transactive(self, *, manager, producer):
        manager.app.assignor._topic_groups = {
            't': 3,
        }
        manager.consumer.key_partition.return_value = None

        await manager.send(
            't', 'k', 'v', partition=None, headers=None, timestamp=None)
        manager.consumer.key_partition.assert_called_once_with('t', 'k', None)
        producer.send.assert_called_once_with(
            't', 'k', 'v', None, None, None, transactional_id=None,
        )

    def test_send_soon(self, *, manager):
        with pytest.raises(NotImplementedError):
            manager.send_soon(Mock(name='FutureMessage'))

    @pytest.mark.asyncio
    async def test_send_and_wait(self, *, manager):
        on_send = Mock()

        async def send(*args, **kwargs):
            on_send(*args, **kwargs)
            return done_future()
        manager.send = send

        await manager.send_and_wait('t', 'k', 'v', 3, 43.2, {})
        on_send.assert_called_once_with(
            't', 'k', 'v', 3, 43.2, {},
        )

    @pytest.mark.asyncio
    async def test_commit(self, *, manager, producer):
        manager.app.assignor._topic_groups = {
            'foo': 1,
            'bar': 2,
        }
        await manager.commit(
            {
                TP('foo', 0): 3003,
                TP('bar', 0): 3004,
                TP('foo', 3): 4004,
                TP('foo', 1): 4005,
            },
            start_new_transaction=False,
        )
        producer.commit_transactions.assert_called_once_with(
            {
                'testid-1-0': {
                    TP('foo', 0): 3003,
                },
                'testid-1-3': {
                    TP('foo', 3): 4004,
                },
                'testid-1-1': {
                    TP('foo', 1): 4005,
                },
                'testid-2-0': {
                    TP('bar', 0): 3004,
                },
            },
            'testid',
            start_new_transaction=False,
        )

    @pytest.mark.asyncio
    async def test_commit__empty(self, *, manager):
        await manager.commit({}, start_new_transaction=False)

    def test_key_partition(self, *, manager):
        with pytest.raises(NotImplementedError):
            manager.key_partition('topic', 'key')

    @pytest.mark.asyncio
    async def test_flush(self, *, manager, producer):
        await manager.flush()
        producer.flush.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_create_topic(self, *, manager):
        await manager.create_topic(
            topic='topic',
            partitions=100,
            replication=3,
            config={'C': 1},
            timeout=30.0,
            retention=40.0,
            compacting=True,
            deleting=True,
            ensure_created=True,
        )
        manager.producer.create_topic.assert_called_once_with(
            'topic', 100, 3,
            config={'C': 1},
            timeout=30.0,
            retention=40.0,
            compacting=True,
            deleting=True,
            ensure_created=True,
        )

    def test_supports_headers(self, *, manager):
        ret = manager.supports_headers()
        assert ret is manager.producer.supports_headers.return_value


class MockedConsumerAbstractMethods:

    def assignment(self):
        return self.current_assignment

    def position(self, *args, **kwargs):
        ...

    async def create_topic(self, *args, **kwargs):
        ...

    def earliest_offsets(self, *args, **kwargs):
        ...

    def highwater(self, *args, **kwargs):
        ...

    def highwaters(self, *args, **kwargs):
        ...

    async def _getmany(self, *args, **kwargs):
        ...

    async def _seek(self, *args, **kwargs):
        ...

    def _to_message(self, *args, **kwargs):
        ...

    async def seek_to_committed(self, *args, **kwargs):
        ...

    async def seek_wait(self, *args, **kwargs):
        ...

    async def subscribe(self, *args, **kwargs):
        ...

    async def seek_to_beginning(self, *args, **kwargs):
        ...

    async def _commit(self, offsets) -> bool:
        ...

    def topic_partitions(self, topic):
        ...

    def _new_topicpartition(self, topic, partition) -> TP:
        return TP(topic, partition)

    def key_partition(self, *args, **kwargs):
        ...


class MyConsumer(MockedConsumerAbstractMethods, Consumer):

    def __init__(self, *args, **kwargs) -> None:
        self.current_assignment = set()
        super().__init__(*args, **kwargs)


class test_Consumer:

    @pytest.fixture
    def callback(self):
        return Mock(name='callback')

    @pytest.fixture
    def on_P_revoked(self):
        return Mock(name='on_partitions_revoked')

    @pytest.fixture
    def on_P_assigned(self):
        return Mock(name='on_partitions_assigned')

    @pytest.fixture
    def consumer(self, *,
                 app,
                 callback,
                 on_P_revoked,
                 on_P_assigned):
        return MyConsumer(
            app.transport,
            callback=callback,
            on_partitions_revoked=on_P_revoked,
            on_partitions_assigned=on_P_assigned,
        )

    @pytest.fixture
    def message(self):
        return Mock(name='message', autospec=Message)

    def test_on_init_dependencies__default(self, *, consumer):
        consumer.in_transaction = False
        assert consumer.on_init_dependencies() == []

    def test_on_init_dependencies__exactly_once(self, *, consumer):
        consumer.in_transaction = True
        assert consumer.on_init_dependencies() == [consumer.transactions]

    @pytest.mark.asyncio
    async def test_getmany__stopped_after_wait(self, *, consumer):
        consumer._wait_next_records = AsyncMock()

        async def on_wait(timeout):
            consumer._stopped.set()
            return None, None

        consumer._wait_next_records.side_effect = on_wait
        assert [a async for a in consumer.getmany(1.0)] == []

    @pytest.mark.asyncio
    async def test_getmany__flow_inactive(self, *, consumer):
        consumer._wait_next_records = AsyncMock(return_value=(
            {TP1: ['A', 'B', 'C']},
            {TP1},
        ))
        consumer.flow_active = False
        assert [a async for a in consumer.getmany(1.0)] == []

    @pytest.mark.asyncio
    async def test_getmany__flow_inactive2(self, *, consumer):
        consumer._wait_next_records = AsyncMock(return_value=(
            {TP1: ['A', 'B', 'C'], TP2: ['D']},
            {TP1},
        ))
        consumer.scheduler = Mock()

        def se(records):
            for value in records.items():
                yield value
                consumer.flow_active = False
        consumer.scheduler.iterate.side_effect = se

        consumer.flow_active = True
        res = [a async for a in consumer.getmany(1.0)]
        assert res
        assert len(res) == 1

    @pytest.mark.asyncio
    async def test_getmany(self, *, consumer):
        def to_message(tp, record):
            return record
        consumer._to_message = to_message
        self._setup_records(
            consumer,
            active_partitions={TP1, TP2},
            records={
                TP1: ['A', 'B', 'C'],
                TP2: ['D', 'E', 'F', 'G'],
                TP3: ['H', 'I', 'J'],
            },
        )
        assert not consumer.should_stop
        consumer.flow_active = False
        consumer.can_resume_flow.set()
        assert [a async for a in consumer.getmany(1.0)] == []
        assert not consumer.should_stop
        consumer.flow_active = True
        assert [a async for a in consumer.getmany(1.0)] == [
            (TP1, 'A'),
            (TP2, 'D'),
            (TP1, 'B'),
            (TP2, 'E'),
            (TP1, 'C'),
            (TP2, 'F'),
            (TP2, 'G'),
        ]

    @pytest.mark.asyncio
    @pytest.mark.parametrize('client_only', [
        False,
        True,
    ])
    async def test__wait_next_records(self, client_only, *, consumer):
        consumer.app.client_only = client_only
        self._setup_records(
            consumer,
            active_partitions={TP1},
            records={
                TP1: ['A', 'B', 'C'],
            },
        )
        expected_active = None if client_only else {TP1}
        ret = await consumer._wait_next_records(1.0)
        assert ret == ({TP1: ['A', 'B', 'C']}, expected_active)

    @pytest.mark.asyncio
    async def test__wait_next_records__flow_inactive(self, *, consumer):
        self._setup_records(consumer, {TP1}, flow_active=False)
        consumer.can_resume_flow.set()
        consumer.wait = AsyncMock()
        await consumer._wait_next_records(1.0)
        consumer.wait.assert_called_once_with(consumer.can_resume_flow)

    @pytest.mark.asyncio
    async def test__wait_next_records__no_active_tps(self, *, consumer):
        self._setup_records(consumer, set())
        consumer.sleep = AsyncMock()
        ret = await consumer._wait_next_records(1.0)
        consumer.sleep.assert_called_once_with(1)
        assert ret == ({}, set())

    def _setup_records(self, consumer, active_partitions,
                       records=None,
                       flow_active=True):
        consumer.flow_active = flow_active
        consumer._active_partitions = active_partitions
        consumer._getmany = AsyncMock(
            return_value={} if records is None else records,
        )

    @pytest.mark.asyncio
    async def test__wait_for_ack(self, *, consumer):

        async def set_ack():
            await asyncio.sleep(0)
            consumer._waiting_for_ack.set_result(None)

        await asyncio.gather(
            consumer._wait_for_ack(timeout=1.0),
            set_ack(),
        )

    @pytest.mark.asyncio
    async def test_on_restart(self, *, consumer):
        consumer._reset_state = Mock()
        consumer.on_init = Mock()
        await consumer.on_restart()
        consumer._reset_state.assert_called_once_with()
        consumer.on_init.assert_called_once_with()

    def test__get_active_partitions__when_empty(self, *, consumer):
        consumer._active_partitions = None
        consumer.assignment = Mock(return_value={TP1})
        assert consumer._get_active_partitions() == {TP1}

    def test__get_active_partitions__when_set(self, *, consumer):
        consumer._active_partitions = {TP1, TP2}
        consumer.assignment = Mock(return_value={TP1})
        assert consumer._get_active_partitions() == {TP1, TP2}

    @pytest.mark.asyncio
    async def test_perform_seek(self, *, consumer):
        consumer._read_offset.update(TP1=3001, TP2=3002)
        consumer._committed_offset.update(TP1=301, TP2=302)
        consumer.seek_to_committed = AsyncMock(return_value={
            TP1: 4001,
            TP2: 0,
        })

        await consumer.perform_seek()

        assert consumer._read_offset[TP1] == 4001
        assert consumer._read_offset[TP2] == 0

        assert consumer._committed_offset[TP1] == 4001
        assert consumer._committed_offset[TP2] is None

    @pytest.mark.asyncio
    async def test_commit__client_only(self, *, consumer):
        consumer.app.client_only = True
        assert not await consumer.commit()

    @pytest.mark.asyncio
    async def test_seek(self, *, consumer):
        consumer._read_offset[TP1] = 301
        consumer._seek = AsyncMock()

        await consumer.seek(TP1, 401)

        assert consumer._read_offset[TP1] == 401
        consumer._seek.assert_called_once_with(TP1, 401)

    def test_stop_flow(self, *, consumer):
        consumer.flow_active = True
        consumer.can_resume_flow.set()

        consumer.stop_flow()

        assert not consumer.flow_active
        assert not consumer.can_resume_flow.is_set()

    def test_resume_flow(self, *, consumer):
        consumer.flow_active = False
        consumer.can_resume_flow.clear()

        consumer.resume_flow()

        assert consumer.flow_active
        assert consumer.can_resume_flow.is_set()

    def test_pause_partitions(self, *, consumer):
        consumer._paused_partitions.clear()
        consumer._active_partitions = {TP1, TP2}

        consumer.pause_partitions([TP2])
        assert consumer._active_partitions == {TP1}
        assert consumer._paused_partitions == {TP2}

    def test_resume_partitions(self, *, consumer):
        consumer._paused_partitions = {TP2}
        consumer._active_partitions = {TP1}

        consumer.resume_partitions([TP2])
        assert consumer._active_partitions == {TP1, TP2}
        assert not consumer._paused_partitions

    def test_read_offset_default(self, *, consumer):
        assert consumer._read_offset[TP1] is None

    def test_committed_offset_default(self, *, consumer):
        assert consumer._committed_offset[TP1] is None

    def test_is_changelog_tp(self, *, app, consumer):
        app.tables = Mock(name='tables', autospec=TableManager)
        app.tables.changelog_topics = {'foo', 'bar'}
        assert consumer._is_changelog_tp(TP('foo', 31))
        assert consumer._is_changelog_tp(TP('bar', 0))
        assert not consumer._is_changelog_tp(TP('baz', 3))

    @pytest.mark.asyncio
    async def test_on_partitions_revoked(self, *, consumer):
        consumer._on_partitions_revoked = AsyncMock(name='opr')
        tps = {TP('foo', 0), TP('bar', 2)}
        await consumer.on_partitions_revoked(tps)

        consumer._on_partitions_revoked.assert_called_once_with(
            tps)

    @pytest.mark.asyncio
    async def test_on_partitions_revoked__updates_active(self, *, consumer):
        consumer._active_partitions = {TP('foo', 0)}
        consumer._on_partitions_revoked = AsyncMock(name='opr')
        tps = {TP('foo', 0), TP('bar', 2)}
        await consumer.on_partitions_revoked(tps)

        consumer._on_partitions_revoked.assert_called_once_with(
            tps)

        assert not consumer._active_partitions

    @pytest.mark.asyncio
    async def test_on_partitions_assigned(self, *, consumer):
        consumer._on_partitions_assigned = AsyncMock(name='opa')
        tps = {TP('foo', 0), TP('bar', 2)}
        await consumer.on_partitions_assigned(tps)

        consumer._on_partitions_assigned.assert_called_once_with(
            tps)

    def test_track_message(self, *, consumer, message):
        consumer._on_message_in = Mock(name='omin')
        consumer.track_message(message)

        assert message in consumer.unacked
        consumer._on_message_in.assert_called_once_with(
            message.tp, message.offset, message)

    @pytest.mark.parametrize('offset', [
        1,
        303,
    ])
    def test_ack(self, offset, *, consumer, message):
        message.acked = False
        consumer.app = Mock(name='app', autospec=App)
        consumer.app.topics.acks_enabled_for.return_value = True
        consumer._committed_offset[message.tp] = 3
        message.offset = offset
        consumer._acked_index[message.tp] = set()
        consumer.ack(message)
        message.acked = False
        consumer.ack(message)

    def test_ack__already_acked(self, *, consumer, message):
        message.acked = True
        consumer.ack(message)

    def test_ack__disabled(self, *, consumer, message, app):
        message.acked = False
        app.topics = Mock(name='app.topics', autospec=Conductor)
        app.topics.acks_enabled_for.return_value = False
        consumer.ack(message)

    @pytest.mark.asyncio
    async def test_wait_empty(self, *, consumer):
        consumer._unacked_messages = {Mock(autospec=Message)}
        consumer._wait_for_ack = AsyncMock()

        def on_commit(start_new_transaction=True):
            for _ in range(10):
                yield
            if consumer.commit.call_count == 3:
                consumer._unacked_messages.clear()
        consumer.commit = AsyncMock(name='commit', side_effect=on_commit)

        await consumer.wait_empty()

    @pytest.mark.asyncio
    async def test_wait_empty__when_stopped(self, *, consumer):
        consumer._stopped.set()
        await consumer.wait_empty()

    @pytest.mark.asyncio
    async def test_on_stop(self, *, consumer):
        consumer.app.conf.stream_wait_empty = False
        await consumer.on_stop()

        with pytest.warns(AlreadyConfiguredWarning):
            consumer.app.conf.stream_wait_empty = True
        consumer.wait_empty = AsyncMock(name='wait_empty')

        await consumer.on_stop()
        consumer.wait_empty.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_force_commit(self, *, consumer):
        consumer.app = Mock(name='app', autospec=App)
        oci = consumer.app.sensors.on_commit_initiated
        occ = consumer.app.sensors.on_commit_completed
        consumer._commit_tps = AsyncMock(name='_commit_tps')
        consumer._acked = {
            TP1: [1, 2, 3, 4, 5],
        }
        consumer._committed_offset = {
            TP1: 2,
        }
        await consumer.force_commit({TP1})
        oci.assert_called_once_with(consumer)
        consumer._commit_tps.assert_called_once_with(
            [TP1],
            start_new_transaction=True,
        )
        occ.assert_called_once_with(consumer, oci())

    @pytest.mark.asyncio
    async def test_commit_tps(self, *, consumer):
        consumer._handle_attached = AsyncMock(name='_handle_attached')
        consumer._commit_offsets = AsyncMock(name='_commit_offsets')
        consumer._filter_committable_offsets = Mock(name='filt')
        consumer._filter_committable_offsets.return_value = {
            TP1: 4,
            TP2: 30,
        }
        await consumer._commit_tps(
            {TP1, TP2},
            start_new_transaction=False,
        )

        consumer._handle_attached.assert_called_once_with({
            TP1: 4,
            TP2: 30,
        })
        consumer._commit_offsets.assert_called_once_with(
            {TP1: 4, TP2: 30},
            start_new_transaction=False,
        )

    @pytest.mark.asyncio
    async def test_commit_tps__ProducerSendError(self, *, consumer):
        consumer._handle_attached = Mock(name='_handle_attached')
        exc = consumer._handle_attached.side_effect = ProducerSendError()
        consumer.crash = AsyncMock(name='crash')
        consumer._filter_committable_offsets = Mock(name='filt')
        consumer._filter_committable_offsets.return_value = {
            TP1: 4,
            TP2: 30,
        }
        await consumer._commit_tps(
            {TP1, TP2},
            start_new_transaction=True,
        )

        consumer.crash.assert_called_once_with(exc)

    @pytest.mark.asyncio
    async def test_commit_tps__no_committable(self, *, consumer):
        consumer._filter_committable_offsets = Mock(name='filt')
        consumer._filter_committable_offsets.return_value = {}
        await consumer._commit_tps(
            {TP1, TP2},
            start_new_transaction=True,
        )

    def test_filter_committable_offsets(self, *, consumer):
        consumer._acked = {
            TP1: SortedSet([1, 2, 3, 4, 7, 8]),
            TP2: SortedSet([30, 31, 32, 33, 34, 35, 36, 40]),
        }
        consumer._committed_offset = {
            TP1: 4,
            TP2: 30,
        }
        assert consumer._filter_committable_offsets({TP1, TP2}) == {
            TP2: 36,
        }

    @pytest.mark.asyncio
    async def test_handle_attached(self, *, consumer):
        consumer.app = Mock(
            name='app',
            autospec=App,
            _attachments=Mock(
                autospec=Attachments,
                publish_for_tp_offset=AsyncMock(),
            ),
            producer=Mock(
                autospec=Service,
                wait_many=AsyncMock(),
            ),
        )
        await consumer._handle_attached({
            TP1: 3003,
            TP2: 6006,
        })
        consumer.app._attachments.publish_for_tp_offset.coro.assert_has_calls([
            call(TP1, 3003),
            call(TP2, 6006),
        ])

        consumer.app.producer.wait_many.coro.assert_called_with(ANY)
        att = consumer.app._attachments
        att.publish_for_tp_offset.coro.return_value = None
        await consumer._handle_attached({
            TP1: 3003,
            TP2: 6006,
        })

    @pytest.mark.asyncio
    async def test_commit_offsets(self, *, consumer):
        consumer._commit = AsyncMock(name='_commit')
        consumer.current_assignment.update({TP1, TP2})
        await consumer._commit_offsets({
            TP1: 3003,
            TP2: 6006,
        })
        consumer._commit.assert_called_once_with({
            TP1: 3003,
            TP2: 6006,
        })

    @pytest.mark.asyncio
    async def test_commit_offsets__did_not_commit(self, *, consumer):
        consumer.in_transaction = False
        consumer._commit = AsyncMock(return_value=False)
        consumer.current_assignment.update({TP1, TP2})
        consumer.app.tables = Mock(name='app.tables')
        await consumer._commit_offsets({
            TP1: 3003,
            TP2: 6006,
            TP3: 7007,
        })
        consumer.app.tables.on_commit.assert_not_called()

    @pytest.mark.asyncio
    async def test_commit_offsets__in_transaction(self, *, consumer):
        consumer.in_transaction = True
        consumer.transactions.commit = AsyncMock()
        consumer.current_assignment.update({TP1, TP2})
        ret = await consumer._commit_offsets({
            TP1: 3003,
            TP2: 6006,
            TP3: 7007,
        })
        consumer.transactions.commit.assert_called_once_with(
            {TP1: 3003, TP2: 6006},
            start_new_transaction=True,
        )
        assert ret is consumer.transactions.commit.coro()

    @pytest.mark.asyncio
    async def test_commit_offsets__no_committable_offsets(self, *, consumer):
        consumer.current_assignment.clear()
        assert not await consumer._commit_offsets({
            TP1: 3003,
            TP2: 6006,
            TP3: 7007,
        })

    @pytest.mark.asyncio
    async def test_commit__already_committing(self, *, consumer):
        consumer.maybe_wait_for_commit_to_finish = AsyncMock(
            return_value=True)
        assert not await consumer.commit()

    @pytest.mark.asyncio
    async def test_commit(self, *, consumer):
        topics = {'foo', 'bar'}
        start_new_transaction = False
        consumer.maybe_wait_for_commit_to_finish = AsyncMock(
            return_value=False)
        consumer.force_commit = AsyncMock()
        ret = await consumer.commit(
            topics,
            start_new_transaction=start_new_transaction,
        )
        consumer.force_commit.assert_called_once_with(
            topics,
            start_new_transaction=start_new_transaction,
        )
        assert ret is consumer.force_commit.coro()
        assert consumer._commit_fut is None

    def test_filter_tps_with_pending_acks(self, *, consumer):
        consumer._acked = {
            TP1: SortedSet([1, 2, 3, 4, 5, 6]),
            TP2: SortedSet([3, 4, 5, 6]),
        }
        assert list(consumer._filter_tps_with_pending_acks()) == [
            TP1, TP2,
        ]
        assert list(consumer._filter_tps_with_pending_acks([TP1])) == [
            TP1,
        ]
        assert list(consumer._filter_tps_with_pending_acks([TP1.topic])) == [
            TP1, TP2,
        ]

    @pytest.mark.parametrize('tp,offset,committed,should', [
        (TP1, 0, 0, False),
        (TP1, 1, 0, True),
        (TP1, 6, 8, False),
        (TP1, 100, 8, True),
    ])
    def test_should_commit(self, tp, offset, committed, should, *, consumer):
        consumer._committed_offset[tp] = committed
        assert consumer._should_commit(tp, offset) == should

    @pytest.mark.parametrize('tp,acked,expected_offset', [
        (TP1, [], None),
        (TP1, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 10),
        (TP1, [1, 2, 3, 4, 5, 6, 7, 8, 10], 8),
        (TP1, [1, 2, 3, 4, 6, 7, 8, 10], 4),
        (TP1, [1, 3, 4, 6, 7, 8, 10], 1),
    ])
    def test_new_offset(self, tp, acked, expected_offset, *, consumer):
        consumer._acked[tp] = SortedSet(acked)
        assert consumer._new_offset(tp) == expected_offset

    @pytest.mark.parametrize('tp,acked,gaps,expected_offset', [
        (TP1, [], [], None),
        (TP1, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], [], 10),
        (TP1, [1, 2, 3, 4, 5, 6, 7, 8, 10], [(9, 10)], 10),
        (TP1, [1, 2, 3, 4, 6, 7, 8, 10], [(5, 6)], 8),
        (TP1, [1, 3, 4, 6, 7, 8, 10], [(2, 3), (5, 6), (9, 10)], 10),
    ])
    def test_new_offset_with_gaps(self, tp, acked, gaps,
                                  expected_offset, *, consumer):
        consumer._acked[tp] = SortedSet(acked)
        tree = IntervalTree()
        for start, end in gaps:
            tree.addi(start, end)
        consumer._gap[tp] = tree
        assert consumer._new_offset(tp) == expected_offset

    @pytest.mark.asyncio
    async def test_on_task_error(self, *, consumer):
        consumer.commit = AsyncMock(name='commit')
        await consumer.on_task_error(KeyError())
        consumer.commit.assert_called_once_with()

    def test__add_gap(self, *, consumer):
        tp = TP1
        consumer._committed_offset[tp] = 299
        consumer._add_gap(TP1, 300, 343)

        assert consumer._gap[tp] == list(range(300, 343))

    def test__add_gap__previous_to_committed(self, *, consumer):
        tp = TP1
        consumer._committed_offset[tp] = 400
        consumer._add_gap(TP1, 300, 343)

        assert consumer._gap[tp] == []

    @pytest.mark.asyncio
    async def test_commit_handler(self, *, consumer):
        i = 0

        def on_sleep(secs, **kwargs):
            nonlocal i
            if i > 1:
                consumer._stopped.set()
            i += 1

        consumer.sleep = AsyncMock(name='sleep', side_effect=on_sleep)
        consumer.commit = AsyncMock(name='commit')

        await consumer._commit_handler(consumer)
        consumer.sleep.coro.assert_has_calls([
            call(consumer.commit_interval),
            call(pytest.approx(consumer.commit_interval, rel=1e-1)),
        ])
        consumer.commit.assert_called_once_with()

    def test_close(self, *, consumer):
        consumer.close()


class test_ConsumerThread:

    class MyConsumerThread(MockedConsumerAbstractMethods, ConsumerThread):

        def close(self):
            ...

        async def getmany(self, *args, **kwargs):
            yield None, None

        def pause_partitions(self, *args, **kwargs):
            ...

        def resume_partitions(self, *args, **kwargs):
            ...

        def stop_flow(self, *args, **kwargs):
            ...

        def resume_flow(self, *args, **kwargs):
            ...

        async def commit(self, *args, **kwargs):
            ...

        async def perform_seek(self, *args, **kwargs):
            ...

        async def seek(self, *args, **kwargs):
            ...

    @pytest.fixture
    def consumer(self):
        return Mock(
            name='consumer',
            spec=Consumer,
            threadsafe_partitions_revoked=AsyncMock(),
            threadsafe_partitions_assigned=AsyncMock(),
            transport=Mock(),
        )

    @pytest.fixture
    def thread(self, *, consumer):
        return self.MyConsumerThread(consumer)

    @pytest.mark.asyncio
    async def test_on_partitions_revoked(self, *, thread, consumer):
        await thread.on_partitions_revoked({TP1, TP2})
        consumer.threadsafe_partitions_revoked.assert_called_once_with(
            thread.thread_loop, {TP1, TP2},
        )

    @pytest.mark.asyncio
    async def test_on_partitions_assigned(self, *, thread, consumer):
        await thread.on_partitions_assigned({TP1, TP2})
        consumer.threadsafe_partitions_assigned.assert_called_once_with(
            thread.thread_loop, {TP1, TP2},
        )


class test_ThreadDelegateConsumer:

    class TestThreadDelegateConsumer(ThreadDelegateConsumer):

        def _new_consumer_thread(self):
            return Mock(
                name='consumer._thread',
                spec=Consumer,
                getmany=AsyncMock(),
                subscribe=AsyncMock(),
                seek_to_committed=AsyncMock(),
                position=AsyncMock(),
                seek_wait=AsyncMock(),
                seek=AsyncMock(),
                earliest_offsets=AsyncMock(),
                highwaters=AsyncMock(),
                commit=AsyncMock(),
            )

        def _new_topicpartition(self, *args, **kwargs):
            return TP(*args, **kwargs)

        def _to_message(self, *args, **kwargs):
            return (args, kwargs)

        def create_topic(self, *args, **kwargs):
            return

    @pytest.fixture
    def message_callback(self):
        return Mock(name='message_callback')

    @pytest.fixture
    def partitions_revoked_callback(self):
        return Mock(name='partitions_revoked_callback')

    @pytest.fixture
    def partitions_assigned_callback(self):
        return Mock(name='partitions_assigned_callback')

    @pytest.fixture
    def consumer(self, *,
                 app,
                 message_callback,
                 partitions_revoked_callback,
                 partitions_assigned_callback):
        consumer = self.TestThreadDelegateConsumer(
            app.transport,
            callback=message_callback,
            on_partitions_revoked=partitions_revoked_callback,
            on_partitions_assigned=partitions_assigned_callback,
        )

        async def _call(*args, **kwargs):
            consumer._method_queue._call(*args, **kwargs)
            return done_future()

        consumer._method_queue = Mock(
            name='_method_queue',
            spec=MethodQueue,
            call=_call,
            _call=Mock(),
        )
        return consumer

    @pytest.mark.asyncio
    async def test_threadsafe_partitions_revoked(self, *, consumer):
        loop = Mock(name='loop')
        await consumer.threadsafe_partitions_revoked(loop, {})
        loop.create_future.assert_called_once_with()
        consumer._method_queue._call.assert_called_once_with(
            loop.create_future(),
            consumer.on_partitions_revoked,
            {},
        )

    @pytest.mark.asyncio
    async def test_threadsafe_partitions_assigned(self, *, consumer):
        loop = Mock(name='loop')
        await consumer.threadsafe_partitions_assigned(loop, {})
        loop.create_future.assert_called_once_with()
        consumer._method_queue._call.assert_called_once_with(
            loop.create_future(),
            consumer.on_partitions_assigned,
            {},
        )

    @pytest.mark.asyncio
    async def test__getmany(self, *, consumer):
        ret = await consumer._getmany({TP1, TP2}, timeout=30.334)
        consumer._thread.getmany.assert_called_once_with(
            {TP1, TP2}, 30.334,
        )
        assert ret is consumer._thread.getmany.coro.return_value

    @pytest.mark.asyncio
    async def test_subscribe(self, *, consumer):
        await consumer.subscribe(topics=['a', 'b', 'c'])
        consumer._thread.subscribe.assert_called_once_with(
            topics=['a', 'b', 'c'],
        )

    @pytest.mark.asyncio
    async def test_seek_to_committed(self, *, consumer):
        ret = await consumer.seek_to_committed()
        consumer._thread.seek_to_committed.assert_called_once_with()
        assert ret is consumer._thread.seek_to_committed.coro.return_value

    @pytest.mark.asyncio
    async def test_position(self, *, consumer):
        ret = await consumer.position(TP1)
        consumer._thread.position.assert_called_once_with(TP1)
        assert ret is consumer._thread.position.coro.return_value

    @pytest.mark.asyncio
    async def test_seek_wait(self, *, consumer):
        await consumer.seek_wait({TP1: 301})
        consumer._thread.seek_wait.assert_called_once_with({TP1: 301})

    @pytest.mark.asyncio
    async def test__seek(self, *, consumer):
        await consumer._seek(TP1, 302)
        consumer._thread.seek.assert_called_once_with(TP1, 302)

    def test_assignment(self, *, consumer):
        ret = consumer.assignment()
        consumer._thread.assignment.assert_called_once_with()
        assert ret is consumer._thread.assignment()

    def test_highwater(self, *, consumer):
        ret = consumer.highwater(TP2)
        consumer._thread.highwater.assert_called_once_with(TP2)
        assert ret is consumer._thread.highwater()

    def test_topic_partitions(self, *, consumer):
        ret = consumer.topic_partitions(TP2.topic)
        consumer._thread.topic_partitions.assert_called_once_with(TP2.topic)
        assert ret is consumer._thread.topic_partitions()

    @pytest.mark.asyncio
    async def test_earliest_offsets(self, *, consumer):
        ret = await consumer.earliest_offsets(TP1, TP2)
        consumer._thread.earliest_offsets.assert_called_once_with(TP1, TP2)
        assert ret is consumer._thread.earliest_offsets.coro.return_value

    @pytest.mark.asyncio
    async def test_highwaters(self, *, consumer):
        ret = await consumer.highwaters(TP1, TP2)
        consumer._thread.highwaters.assert_called_once_with(TP1, TP2)
        assert ret is consumer._thread.highwaters.coro.return_value

    @pytest.mark.asyncio
    async def test_commit(self, *, consumer):
        ret = await consumer._commit({TP1: 301, TP2: 302})
        consumer._thread.commit.assert_called_once_with(
            {TP1: 301, TP2: 302},
        )
        assert ret is consumer._thread.commit.coro.return_value

    @pytest.mark.asyncio
    async def test_maybe_wait_for_commit_to_finish(self, *, loop, consumer):
        consumer._commit_fut = None
        assert not await consumer.maybe_wait_for_commit_to_finish()
        consumer._commit_fut = loop.create_future()
        consumer._commit_fut.cancel()
        assert not await consumer.maybe_wait_for_commit_to_finish()
        consumer._commit_fut = loop.create_future()
        consumer._commit_fut.set_result(None)
        assert await consumer.maybe_wait_for_commit_to_finish()

    @pytest.mark.asyncio
    async def test_close(self, *, consumer):
        consumer.close()
        consumer._thread.close.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_key_partition(self, *, consumer):
        ret = consumer.key_partition('topic', 'key', partition=3)
        consumer._thread.key_partition.assert_called_once_with(
            'topic', 'key', partition=3,
        )
        assert ret is consumer._thread.key_partition()

    def test_verify_event_path(self, *, consumer):
        assert consumer.verify_event_path(30.1, TP1) is None

    @pytest.mark.asyncio
    async def test_verify_all_partitions_active(self, *, consumer):
        consumer.assignment = Mock(name='assignment')
        # use list for order, argument is originally a set.
        consumer.assignment.return_value = [TP1, TP2, TP3]

        consumer.verify_event_path = Mock(name='verify_event_path')

        with patch('faust.transport.consumer.monotonic') as monotonic:
            now = monotonic.return_value = 391243.231
            await consumer.verify_all_partitions_active()

            consumer.verify_event_path.assert_has_calls([
                call(now, TP1),
                call(now, TP2),
                call(now, TP3),
            ])

    @pytest.mark.asyncio
    async def test_verify_all_partitions_active__bail_on_sleep(
            self, *, consumer):
        consumer.assignment = Mock(name='assignment')
        consumer.assignment.return_value = [TP1, TP2, TP3]  # must be ordered

        consumer.verify_event_path = Mock(name='verify_event_path')
        consumer.sleep = AsyncMock()

        async def on_sleep(secs):
            if consumer.sleep.call_count == 2:
                consumer._stopped.set()

        consumer.sleep.side_effect = on_sleep

        with patch('faust.transport.consumer.monotonic') as monotonic:
            now = monotonic.return_value = 391243.231
            await consumer.verify_all_partitions_active()

            consumer.verify_event_path.assert_called_once_with(now, TP1)



import abc
from datetime import datetime
from decimal import Decimal
from typing import ClassVar, Dict, List, Mapping, Optional, Set, Tuple
from dateutil.parser import parse as parse_date
import faust
from faust.exceptions import SecurityError, ValidationError
from faust.models import maybe_model
from faust.models.fields import (
    BytesField,
    DatetimeField,
    DecimalField,
    FieldDescriptor,
    FloatField,
    IntegerField,
    StringField,
)
from mode.utils.logging import get_logger
from faust.models.tags import (
    Secret,
    Sensitive,
    _FrameLocal,
    allow_protected_vars,
)
from faust.types import ModelT
from faust.utils import json
import pytest

logger = get_logger(__name__)


class Record(faust.Record, serializer='json'):
    # make sure all models in this test uses json as the serializer.
    ...


class Account(Record):
    id: str
    name: str
    active: bool = True


class AccountList(Record):
    accounts: List[Account]


class AccountSet(Record):
    accounts: Set[Account]


class AccountMap(Record):
    accounts: Dict[str, Account]


class FREFAccountList(Record):
    accounts: 'List[Account]'


class FREFAccountSet(Record):
    accounts: 'Set[Account]'


class FREFAccountMap(Record):
    accounts: Dict[str, 'Account']
    something: 'ClassVar[bool]' = True


class User(Record):
    id: str
    username: str
    account: Account


def test_parameters():
    account = Account('id', 'name', True)
    assert account.id == 'id'
    assert account.name == 'name'
    assert account.active

    account2 = Account('id', 'name', active=False)
    assert account2.id == 'id'
    assert account2.name == 'name'
    assert not account2.active

    class Account3(Account):
        foo: int = None

    account3 = Account3('id', 'name', False, 'foo')
    assert account3.id == 'id'
    assert account3.name == 'name'
    assert not account3.active
    assert account3.foo == 'foo'

    with pytest.raises(AttributeError):
        account2.foo


def test_paramters_with_custom_init():

    class Point(Record, include_metadata=False):
        x: int
        y: int

        def __init__(self, x, y, **kwargs):
            self.x = x
            self.y = y

    p = Point(30, 10)
    assert p.x == 30
    assert p.y == 10

    payload = p.dumps(serializer='json')
    assert payload == b'{"x": 30, "y": 10}'

    data = json.loads(payload)
    p2 = Point.from_data(data)
    assert p2.x == 30
    assert p2.y == 10


def test_parameters_with_custom_init_and_super():

    class Point(Record, include_metadata=False):
        x: int
        y: int

        def __post_init__(self):
            self.z = self.x + self.y

    p = Point(30, 10)
    assert p.x == 30
    assert p.y == 10
    assert p.z == 40

    payload = p.dumps(serializer='json')
    assert payload == b'{"x": 30, "y": 10}'

    data = json.loads(payload)
    p2 = Point.from_data(data)
    assert p2.x == 30
    assert p2.y == 10
    assert p2.z == 40


def test_datetimes():

    class Date(Record, coerce=True):
        date: datetime

    class OptionalDate(Record, coerce=True):
        date: Optional[datetime]

    class TupleOfDate(Record, coerce=True):
        dates: Tuple[datetime, ...]

    class SetOfDate(Record, coerce=True):
        dates: Set[datetime]

    class MapOfDate(Record, coerce=True):
        dates: Mapping[int, datetime]

    class ListOfDate(Record, coerce=True):
        dates: List[datetime]

    class OptionalListOfDate(Record, coerce=True):
        dates: List[datetime] = None

    class OptionalListOfDate2(Record, coerce=True):
        dates: Optional[List[datetime]]

    assert isinstance(OptionalDate.date, DatetimeField)

    n1 = datetime.utcnow()
    assert Date.loads(Date(date=n1).dumps()).date == n1
    assert OptionalDate.loads(OptionalDate(date=n1).dumps()).date == n1
    assert OptionalDate.loads(OptionalDate(date=None).dumps()).date is None
    n2 = datetime.utcnow()
    assert ListOfDate.loads(ListOfDate(
        dates=[n1, n2]).dumps()).dates == [n1, n2]
    assert OptionalListOfDate.loads(OptionalListOfDate(
        dates=None).dumps()).dates is None
    assert OptionalListOfDate.loads(OptionalListOfDate(
        dates=[n2, n1]).dumps()).dates == [n2, n1]
    assert OptionalListOfDate2.loads(OptionalListOfDate2(
        dates=None).dumps()).dates is None
    assert OptionalListOfDate2.loads(OptionalListOfDate2(
        dates=[n1, n2]).dumps()).dates == [n1, n2]
    assert TupleOfDate.loads(TupleOfDate(
        dates=(n1, n2)).dumps()).dates == (n1, n2)
    assert TupleOfDate.loads(TupleOfDate(
        dates=(n2,)).dumps()).dates == (n2,)
    assert SetOfDate.loads(SetOfDate(
        dates={n1, n2}).dumps()).dates == {n1, n2}
    assert MapOfDate.loads(MapOfDate(
        dates={'A': n1, 'B': n2}).dumps()).dates == {'A': n1, 'B': n2}

    datelist = ListOfDate.from_data({
        'dates': [n1.isoformat(), n2.isoformat()]})
    assert isinstance(datelist.dates[0], datetime)
    assert isinstance(datelist.dates[1], datetime)


def test_datetimes__isodates_compat():

    class Date(Record, coerce=False, isodates=True):
        date: datetime

    class TupleOfDate(Record, coerce=False, isodates=True):
        dates: Tuple[datetime, ...]

    class SetOfDate(Record, coerce=False, isodates=True):
        dates: Set[datetime]

    class MapOfDate(Record, coerce=False, isodates=True):
        dates: Mapping[int, datetime]

    class ListOfDate(Record, coerce=False, isodates=True):
        dates: List[datetime]

    class OptionalListOfDate(Record, coerce=False, isodates=True):
        dates: List[datetime] = None

    class OptionalListOfDate2(Record, coerce=False, isodates=True):
        dates: Optional[List[datetime]]

    n1 = datetime.utcnow()
    assert Date.loads(Date(date=n1).dumps()).date == n1
    n2 = datetime.utcnow()
    assert ListOfDate.loads(ListOfDate(
        dates=[n1, n2]).dumps()).dates == [n1, n2]
    assert OptionalListOfDate.loads(OptionalListOfDate(
        dates=None).dumps()).dates is None
    assert OptionalListOfDate.loads(OptionalListOfDate(
        dates=[n2, n1]).dumps()).dates == [n2, n1]
    assert OptionalListOfDate2.loads(OptionalListOfDate2(
        dates=None).dumps()).dates is None
    assert OptionalListOfDate2.loads(OptionalListOfDate2(
        dates=[n1, n2]).dumps()).dates == [n1, n2]
    assert TupleOfDate.loads(TupleOfDate(
        dates=(n1, n2)).dumps()).dates == (n1, n2)
    assert TupleOfDate.loads(TupleOfDate(
        dates=(n2,)).dumps()).dates == (n2,)
    assert SetOfDate.loads(SetOfDate(
        dates={n1, n2}).dumps()).dates == {n1, n2}
    assert MapOfDate.loads(MapOfDate(
        dates={'A': n1, 'B': n2}).dumps()).dates == {'A': n1, 'B': n2}

    datelist = ListOfDate.from_data({
        'dates': [n1.isoformat(), n2.isoformat()]})
    assert isinstance(datelist.dates[0], datetime)
    assert isinstance(datelist.dates[1], datetime)


def test_decimals():

    class IsDecimal(Record, coerce=True, serializer='json'):
        number: Decimal

    class ListOfDecimal(Record, coerce=True, serializer='json'):
        numbers: List[Decimal]

    class OptionalListOfDecimal(Record, coerce=True, serializer='json'):
        numbers: List[Decimal] = None

    class OptionalListOfDecimal2(Record, coerce=True, serializer='json'):
        numbers: Optional[List[Decimal]]

    class TupleOfDecimal(Record, coerce=True, serializer='json'):
        numbers: Tuple[Decimal, ...]

    class SetOfDecimal(Record, coerce=True, serializer='json'):
        numbers: Set[Decimal]

    class MapOfDecimal(Record, coerce=True, serializer='json'):
        numbers: Mapping[str, Decimal]

    n1 = Decimal('1.31341324')
    assert IsDecimal.loads(IsDecimal(number=n1).dumps()).number == n1
    n2 = Decimal('3.41569')
    assert ListOfDecimal.loads(ListOfDecimal(
        numbers=[n1, n2]).dumps()).numbers == [n1, n2]
    assert OptionalListOfDecimal.loads(OptionalListOfDecimal(
        numbers=None).dumps()).numbers is None
    assert OptionalListOfDecimal.loads(OptionalListOfDecimal(
        numbers=[n2, n1]).dumps()).numbers == [n2, n1]
    assert OptionalListOfDecimal2.loads(OptionalListOfDecimal2(
        numbers=None).dumps()).numbers is None
    assert OptionalListOfDecimal2.loads(OptionalListOfDecimal2(
        numbers=[n1, n2]).dumps()).numbers == [n1, n2]
    assert TupleOfDecimal.loads(TupleOfDecimal(
        numbers=(n1, n2)).dumps()).numbers == (n1, n2)
    assert TupleOfDecimal.loads(TupleOfDecimal(
        numbers=(n2,)).dumps()).numbers == (n2,)
    assert SetOfDecimal.loads(SetOfDecimal(
        numbers={n1, n2}).dumps()).numbers == {n1, n2}
    assert MapOfDecimal.loads(MapOfDecimal(
        numbers={'A': n1, 'B': n2}).dumps()).numbers == {'A': n1, 'B': n2}

    dlist = ListOfDecimal.from_data({'numbers': ['1.312341', '3.41569']})
    assert isinstance(dlist.numbers[0], Decimal)
    assert isinstance(dlist.numbers[1], Decimal)


def test_decimals_compat():

    class IsDecimal(Record, coerce=False, decimals=True, serializer='json'):
        number: Decimal

    class ListOfDecimal(Record,
                        coerce=False,
                        decimals=True,
                        serializer='json'):
        numbers: List[Decimal]

    class OptionalListOfDecimal(Record,
                                coerce=False,
                                decimals=True,
                                serializer='json'):
        numbers: List[Decimal] = None

    class OptionalListOfDecimal2(Record,
                                 coerce=False,
                                 decimals=True,
                                 serializer='json'):
        numbers: Optional[List[Decimal]]

    class TupleOfDecimal(Record,
                         coerce=False,
                         decimals=True,
                         serializer='json'):
        numbers: Tuple[Decimal, ...]

    class SetOfDecimal(Record,
                       coerce=False,
                       decimals=True,
                       serializer='json'):
        numbers: Set[Decimal]

    class MapOfDecimal(Record,
                       coerce=False,
                       decimals=True,
                       serializer='json'):
        numbers: Mapping[str, Decimal]

    n1 = Decimal('1.31341324')
    assert IsDecimal.loads(IsDecimal(number=n1).dumps()).number == n1
    n2 = Decimal('3.41569')
    assert ListOfDecimal.loads(ListOfDecimal(
        numbers=[n1, n2]).dumps()).numbers == [n1, n2]
    assert OptionalListOfDecimal.loads(OptionalListOfDecimal(
        numbers=None).dumps()).numbers is None
    assert OptionalListOfDecimal.loads(OptionalListOfDecimal(
        numbers=[n2, n1]).dumps()).numbers == [n2, n1]
    assert OptionalListOfDecimal2.loads(OptionalListOfDecimal2(
        numbers=None).dumps()).numbers is None
    assert OptionalListOfDecimal2.loads(OptionalListOfDecimal2(
        numbers=[n1, n2]).dumps()).numbers == [n1, n2]
    assert TupleOfDecimal.loads(TupleOfDecimal(
        numbers=(n1, n2)).dumps()).numbers == (n1, n2)
    assert TupleOfDecimal.loads(TupleOfDecimal(
        numbers=(n2,)).dumps()).numbers == (n2,)
    assert SetOfDecimal.loads(SetOfDecimal(
        numbers={n1, n2}).dumps()).numbers == {n1, n2}
    assert MapOfDecimal.loads(MapOfDecimal(
        numbers={'A': n1, 'B': n2}).dumps()).numbers == {'A': n1, 'B': n2}

    dlist = ListOfDecimal.from_data({'numbers': ['1.312341', '3.41569']})
    assert isinstance(dlist.numbers[0], Decimal)
    assert isinstance(dlist.numbers[1], Decimal)


def test_custom_coercion():

    class Foo:

        def __init__(self, value: int):
            assert isinstance(value, int)
            self.value: int = value

        def __eq__(self, other):
            if other.__class__ is self.__class__:
                return other.value == self.value
            return NotImplemented

        def __hash__(self):
            return hash(self.value)

        def __json__(self):
            return self.value

        def __repr__(self):
            return f'<{type(self).__name__}: {self.value}>'

    class CanFooModel(Record,
                      abstract=True,
                      coercions={Foo: Foo},
                      serializer='json'):
        ...

    class IsFoo(CanFooModel, serializer='json'):
        foo: Foo

    class ListOfFoo(CanFooModel, serializer='json'):
        foos: List[Foo]

    class OptionalListOfFoo(CanFooModel, serializer='json'):
        foos: List[Foo] = None

    class OptionalListOfFoo2(CanFooModel):
        foos: Optional[List[Foo]]

    class TupleOfFoo(CanFooModel):
        foos: Tuple[Foo, ...]

    class SetOfFoo(CanFooModel):
        foos: Set[Foo]

    class MapOfFoo(CanFooModel):
        foos: Mapping[str, Foo]

    n1 = Foo(101)
    assert IsFoo.loads(IsFoo(foo=n1).dumps()).foo == n1
    n2 = Foo(202)
    assert ListOfFoo.loads(ListOfFoo(
        foos=[n1, n2]).dumps()).foos == [n1, n2]
    assert OptionalListOfFoo.loads(OptionalListOfFoo(
        foos=None).dumps()).foos is None
    assert OptionalListOfFoo.loads(OptionalListOfFoo(
        foos=[n2, n1]).dumps()).foos == [n2, n1]
    assert OptionalListOfFoo2.loads(OptionalListOfFoo2(
        foos=None).dumps()).foos is None
    assert OptionalListOfFoo2.loads(OptionalListOfFoo2(
        foos=[n1, n2]).dumps()).foos == [n1, n2]
    assert TupleOfFoo.loads(TupleOfFoo(
        foos=(n1, n2)).dumps()).foos == (n1, n2)
    assert TupleOfFoo.loads(TupleOfFoo(
        foos=(n2,)).dumps()).foos == (n2,)
    assert SetOfFoo.loads(SetOfFoo(
        foos={n1, n2}).dumps()).foos == {n1, n2}
    assert MapOfFoo.loads(MapOfFoo(
        foos={'A': n1, 'B': n2}).dumps()).foos == {'A': n1, 'B': n2}


def test_constructor():
    with pytest.raises(TypeError):
        Account(id='123')
    with pytest.raises(TypeError):
        Account(name='foo')
    with pytest.raises(TypeError):
        Account(unknown_argument=303)
    account = Account(id='123', name='foo')
    assert account.id == '123'
    assert account.name == 'foo'
    assert account.active
    assert not Account(id='123', name='foo', active=False).active


def test_submodels():
    a1 = Account(id='123', name='foo', active=True)
    a2 = Account(id='456', name='bar', active=False)
    a3 = Account(id='789', name='baz', active=True)

    assert AccountList.loads(AccountList(
        accounts=[a1, a2, a3]).dumps()).accounts == [a1, a2, a3]

    expected_set = {a1, a2, a3}
    acc1 = AccountSet.loads(AccountSet(accounts={a1, a2, a3}).dumps()).accounts
    assert isinstance(acc1, set)
    assert len(acc1) == len(expected_set)
    for acc in acc1:
        assert acc in expected_set
    assert AccountMap.loads(AccountMap(
        accounts={'a': a1, 'b': a2, 'c': a3}).dumps()).accounts == {
        'a': a1,
        'b': a2,
        'c': a3,
    }


def test_submodels_forward_reference():
    a1 = Account(id='123', name='foo', active=True)
    a2 = Account(id='456', name='bar', active=False)
    a3 = Account(id='789', name='baz', active=True)

    assert AccountList.loads(FREFAccountList(
        accounts=[a1, a2, a3]).dumps()).accounts == [a1, a2, a3]
    assert sorted(AccountSet.loads(FREFAccountSet(
        accounts={a1, a2, a3}).dumps()).accounts) == sorted({a1, a2, a3})
    assert AccountMap.loads(FREFAccountMap(
        accounts={'a': a1, 'b': a2, 'c': a3}).dumps()).accounts == {
        'a': a1,
        'b': a2,
        'c': a3,
    }


def test_derive():
    a1 = Account(id='123', name='foo', active=True)
    b1 = Account(id='456', name='bar', active=False)
    c1 = Account(id='789', name='baz', active=True)

    assert a1.active
    a2 = a1.derive(active=False)
    assert a2.id == '123'
    assert a2.name == 'foo'
    assert not a2.active

    c2 = a1.derive(b1, c1, name='xuzzy')
    assert c2.id == '789'
    assert c2.name == 'xuzzy'
    assert c2.active

    b2 = b1.derive(active=True)
    assert b2.active
    assert b2.id == '456'
    assert b2.name == 'bar'


def test_classvar_is_not_a_field():

    class PP(Record):
        x: int
        y: int
        z: ClassVar[int] = 3

    p = PP(10, 3)
    assert 'z' not in repr(p)
    assert p.asdict() == {'x': 10, 'y': 3}
    assert p.z == 3

    with pytest.raises(TypeError):
        PP(10, 3, z=300)

    p.z = 4
    assert p.z == 4


def test_constructor_from_data():
    with pytest.raises(TypeError):
        Account.from_data({'id': '123'})
    with pytest.raises(TypeError):
        Account.from_data({'name': 'foo'})
    with pytest.raises(TypeError):
        Account.from_data({'unknown_argument': 303})
    account = Account.from_data({'id': '123', 'name': 'foo'})
    assert account.id == '123'
    assert account.name == 'foo'
    assert account.active
    assert not Account.from_data(
        {'id': '123', 'name': 'foo', 'active': False}).active


@pytest.mark.parametrize('a,b', [
    (User(id=1, username=2, account=Account(id=1, name=2)),
     User(id=1, username=2, account=Account(id=1, name=2))),
    (User(id=1, username=2, account=Account(id=1, name=2, active=False)),
     User(id=1, username=2, account=Account(id=1, name=2, active=False))),
])
def test_eq(a, b):
    assert a == b


def test_eq__incompatible():
    assert Account(id=1, name=2) != object()


@pytest.mark.parametrize('a,b', [
    (User(id=1, username=2, account=Account(id=1, name=2)),
     User(id=2, username=2, account=Account(id=2, name=2))),
    (User(id=1, username=2, account=Account(id=1, name=2)),
     User(id=1, username=3, account=Account(id=1, name=2))),
    (User(id=1, username=2, account=Account(id=1, name=2)),
     User(id=1, username=2, account=Account(id=1, name=3))),
    (User(id=1,
          username=2,
          account=Account(id=1, name=2, active=False)),
     User(id=1,
          username=2,
          account=Account(id=1, name=3, active=True))),
])
def test_ne(a, b):
    assert a != b


def test_json():
    account = Account(id=1, name=2)
    user = User(1, 2, account)

    payload = json.dumps(user)
    deser = json.loads(payload)
    assert deser == {
        'id': 1,
        'username': 2,
        'account': {
            'id': 1,
            'name': 2,
            'active': True,
            '__faust': {'ns': Account._options.namespace},
        },
        '__faust': {'ns': User._options.namespace},
    }

    assert user.__json__() == {
        'id': 1,
        'username': 2,
        'account': account,
        '__faust': {'ns': User._options.namespace},
    }

    assert User.from_data(deser) == user


class test_FieldDescriptor:

    def test_getattr(self):
        u = User(id=1, username=2, account=Account(id=3, name=4))

        assert User.id.getattr(u) == 1
        assert User.username.getattr(u) == 2
        assert User.account.id.getattr(u) == 3
        assert User.account.name.getattr(u) == 4


@pytest.mark.parametrize('record', [
    Account(id=None, name=None),
    Account(id='123', name='123'),
    Account(id='123', name='123', active=False),
    User(id='123',
         username='foo',
         account=Account(id='123', name='Foo', active=True)),
    User(id='123', username='foo', account=None),
    User(id=None, username=None, account=None),
    AccountList(accounts=[Account(id=None, name=None)]),
    AccountMap(accounts={'foo': Account(id=None, name='foo')}),
])
def test_dumps(record):
    assert record.loads(record.dumps()) == record
    assert repr(record)


def test_subclass_default_values():

    class X(Record):
        x: int
        y: int = None

    class Z(X):
        z: int = None

    assert X(x=None).y is None
    assert X(x=None, y=303).y == 303
    assert Z(x=None).y is None
    assert Z(x=None).z is None
    assert Z(x=None, y=101, z=303).y == 101
    assert Z(x=None, y=101, z=303).z == 303


def test_subclass_preserves_required_values():

    class X(Record):
        x: int
        z: int = None

    class Y(X):
        y: int

    assert Y._options.defaults == {'z': None}
    assert Y._options.optionalset == {'z'}
    with pytest.raises(TypeError):
        Y(y=10)
    with pytest.raises(TypeError):
        Y(y=10, z=10)
    Y(x=10, y=20)


def test_too_many_arguments_raises_TypeError():
    class X(Record):
        x: int

    class Y(X):
        y: int

    with pytest.raises(TypeError) as einfo:
        Y(10, 20, 30)
    reason = str(einfo.value)
    assert '__init__() takes' in reason


def test_fields_with_concrete_polymorphic_type__dict():

    class X(Record, isodates=True):
        foo: int
        details: dict


def test_fields_with_concrete_polymorphic_type__dict_optional():

    class X(Record, isodates=True):
        username: str
        metric_name: str
        value: float
        timestamp: datetime
        details: dict = None


def test_fields_with_concrete_polymorphic_type__tuple():

    class X(Record, isodates=True):
        foo: int
        details: tuple


def test_fields_with_concrete_polymorphic_type__list():

    class X(Record, isodates=True):
        foo: int
        details: list


def test_fields_with_concrete_polymorphic_type__set():

    class X(Record, isodates=True):
        foo: int
        details: set


def test_fields_with_concrete_polymorphic_type__frozenset():

    class X(Record, isodates=True):
        foo: int
        details: frozenset


def test_supports_post_init():

    class X(Record):
        x: int
        y: int

        def __post_init__(self):
            self.z: int = self.x + self.y

    x = X(1, 3)
    assert x.z == 4


def test_default_no_blessed_key():

    class X(Record):
        a: int

    class LooksLikeX(Record):
        a: int

    class Y(Record):
        x: X

    x = LooksLikeX(303)
    y = Y(x)

    data = Y.dumps(y, serializer='json')
    y2 = Y.loads(data, serializer='json')
    assert isinstance(y2.x, X)


def test_default_multiple_levels_no_blessed_key():

    class StdAttribution(Record):
        first_name: str
        last_name: str

    class Address(Record):
        country: str

    class Account(StdAttribution):
        address: Address

    class Event(Record):
        account: Account

    event = Event(account=Account(
        first_name='George',
        last_name='Costanza',
        address=Address('US'),
    ))
    s = event.loads(event.dumps(serializer='json'), serializer='json')
    assert isinstance(s.account, Account)
    assert isinstance(s.account.address, Address)


def test_polymorphic_fields(app):

    class X(Record):
        a: int

    class LooksLikeX(Record, polymorphic_fields=True):
        a: int

    class Y(Record):
        x: LooksLikeX

    x = LooksLikeX(303)
    y = Y(x)

    data = Y.dumps(y, serializer='json')
    y2 = app.serializers.loads_key(Y, data, serializer='json')
    assert isinstance(y2.x, LooksLikeX)


def test_compat_enabled_blessed_key(app):

    class X(Record):
        a: int

    class LooksLikeX(Record, allow_blessed_key=True):
        a: int

    class Y(Record):
        x: LooksLikeX

    x = LooksLikeX(303)
    y = Y(x)

    data = Y.dumps(y, serializer='json')
    y2 = app.serializers.loads_key(Y, data, serializer='json')
    assert isinstance(y2.x, LooksLikeX)


def test__polymorphic_fields_deeply_nested():

    class BaseAttribution(Record, abc.ABC):

        def __post_init__(self, *args, **kwargs) -> None:
            self.data_store = None

    class AdjustData(Record):
        activity_kind: str

    class Event(Record):
        category: str
        event: str
        data: AdjustData

    class AdjustRecord(BaseAttribution):
        event: Event

    x = AdjustRecord(Event(
        category='foo',
        event='bar',
        data=AdjustData('baz'),
    ))
    value = x.dumps(serializer='json')
    value_dict = json.loads(value)
    value_dict['event']['__faust']['ns'] = 'x.y.z'
    model = AdjustRecord.from_data(value_dict)
    assert isinstance(model.event, Event)
    assert isinstance(model.event.data, AdjustData)


def test_compat_blessed_key_deeply_nested():

    class BaseAttribution(Record, abc.ABC):

        def __post_init__(self, *args, **kwargs) -> None:
            self.data_store = None

    class AdjustData(Record):
        activity_kind: str

    class Event(Record):
        category: str
        event: str
        data: AdjustData

    class AdjustRecord(BaseAttribution):
        event: Event

    x = AdjustRecord(Event(
        category='foo',
        event='bar',
        data=AdjustData('baz'),
    ))
    value = x.dumps(serializer='json')
    value_dict = json.loads(value)
    value_dict['event']['__faust']['ns'] = 'x.y.z'
    model = AdjustRecord.from_data(value_dict)
    assert isinstance(model.event, Event)
    assert isinstance(model.event.data, AdjustData)


ADTRIBUTE_PAYLOAD = '''
{"user": {"username": "3da6ef8f-aed1-47e7-ad4f-034363d0565b",
 "secret": null, "__faust": {"ns": "trebuchet.models.logging.User"}},
 "device": {"platform": "iOS",
 "device_id": "F5FA74CE-0F17-491F-A6B1-AAE1D036CBF2",
 "os_version": "11.2.6", "manufacturer": "phone",
 "device_version": "iPhone8,2", "screen_resolution": null,
 "source": null, "campaign": null,
 "campaign_version": null, "adid": null,
 "engagement_time": null,
 "__faust": {"ns": "trebuchet.models.logging.Device"}},
 "app": {"version": "4667", "app_id": "com.robinhood.release.Robinhood",
 "build_num": null, "locale": null, "language": null,
 "__faust": {"ns": "trebuchet.models.logging.App"}},
 "event": {"category": "adjust_data", "event": "attribution",
 "experiments": null, "session_id": null, "data":
 {"activity_kind": "session", "network_name": "Organic",
 "adid": "04df21c2ef05a91598e13c82096d921b",
 "tracker": "494pkq", "reftag": "oJuX55u4N4OI4",
 "nonce": "mt5lyv18d", "campaign_name": "",
 "adgroup_name": "", "creative_name": "", "click_referer": "",
 "is_organic": "1", "reattribution_attribution_window": "",
 "impression_attribution_window": "", "store": "itunes",
 "match_type": "", "platform_adid": "", "search_term": "",
 "event_name": "", "installed_at": "2017-09-02 00:28:03.000",
 "engagement_time": null, "deeplink": "",
 "source_user": "", "__faust": {
 "ns": "trebuchet.models.logging_data.AdjustData"}},
 "__faust": {"ns": "trebuchet.models.logging.Event"}},
 "timestamp": "2018-03-22 16:57:19.000", "client_ip": "174.207.10.101",
 "event_hash": "50c9a0e19b9644abe269aadcea9e7526", "__faust": {
    "ns": "trebuchet.models.logging.LoggingEvent"}}
'''


def test_adtribute_payload(app):

    class BaseAttribution(Record, abc.ABC):

        def __post_init__(self) -> None:
            self.data_store = None

    class AdjustData(Record):

        activity_kind: str
        network_name: str
        adid: str
        tracker: str
        reftag: str
        nonce: str
        campaign_name: str = None
        adgroup_name: str = None
        creative_name: str = None
        click_referer: str = None
        is_organic: str = None
        reattribution_attribution_window: str = None
        impression_attribution_window: str = None
        store: str = None
        match_type: str = None
        platform_adid: str = None
        search_term: str = None
        event_name: str = None
        installed_at: str = None
        engagement_time: str = None
        deeplink: str = None
        source_user: str = None

    class User(Record):
        username: str

    class App(Record):
        version: str
        app_id: str

    class Device(Record):
        platform: str
        device_id: str
        os_version: str
        device_version: str
        manufacturer: str

    class Event(Record):
        category: str
        event: str
        data: AdjustData

    class AdjustRecord(BaseAttribution):
        user: User
        device: Device
        app: App
        event: Event
        timestamp: str
        client_ip: str = None
        event_hash: str = None

    def __post_init__(self) -> None:
        self.data_store = None

    app.serializers.loads_value(
        AdjustRecord, ADTRIBUTE_PAYLOAD, serializer='json')


def test_overwrite_asdict():

    with pytest.raises(RuntimeError):

        class R(Record):

            def asdict(self):
                return {'foo': 1}


def test_prepare_dict():

    class Quote(Record):
        ask_price: float = None
        bid_price: float = None

        def _prepare_dict(self, payload):
            return {k: v for k, v in payload.items() if v is not None}

    assert Quote().asdict() == {}
    assert Quote(1.0, 2.0).asdict() == {'ask_price': 1.0, 'bid_price': 2.0}
    assert Quote(None, 2.0).asdict() == {'bid_price': 2.0}


def test_custom_init_calling_model_init():

    class Quote(Record):
        ask_price: float
        bid_price: float

        def __init__(self, ask_price: str, bid_price: str, **kwargs):
            self._model_init(ask_price, bid_price, **kwargs)

    q1 = Quote(1.0, 2.0)
    assert q1.ask_price == 1.0
    assert q1.bid_price == 2.0

    with pytest.raises(TypeError):
        Quote(1.0, 2.0, foo=1)


def test_repr():
    assert repr(Account.id)


def test_ident():
    assert Account.id.ident == 'Account.id'


def test_list_field_refers_to_self():

    class X(Record):
        id: int
        xs: List['X']

    x = X(1, [X(2, [X(3, [])])])

    as_json = x.dumps(serializer='json')
    loads = X.loads(as_json, serializer='json')
    assert loads == x

    assert isinstance(loads.xs[0], X)
    assert isinstance(loads.xs[0].xs[0], X)


def test_optional_modelfield():

    class X(Record):
        id: int

    class Y(Record):
        x: Optional[X] = None

    y = Y(X(30))

    as_json = y.dumps(serializer='json')
    loads = Y.loads(as_json, serializer='json')
    assert loads == y

    assert isinstance(loads.x, X)


def test_optional_modelfield_with_coercion():
    class X(Record, coercions={str: str}):
        y: Optional[str]

    x = X(y='test')

    assert x.y == 'test'


@pytest.mark.parametrize('flag,expected_default', [
    ('isodates', False),
    ('include_metadata', True),
    ('polymorphic_fields', False),
])
def test_subclass_inherit_flags(flag, expected_default):

    class BaseX(Record):
        x: datetime

    X = type('X', (BaseX,), {}, **{flag: not expected_default})
    Y = type('Y', (X,), {})
    Z = type('Z', (Y,), {}, **{flag: expected_default})

    assert getattr(BaseX._options, flag) is expected_default
    assert getattr(X._options, flag) is not expected_default
    assert getattr(Y._options, flag) is not expected_default
    assert getattr(Z._options, flag) is expected_default


def test_abstract_model_repr():

    class MyBase(faust.Record, abstract=True):
        ...

    assert MyBase.__is_abstract__
    with pytest.raises(NotImplementedError):
        MyBase()


def test_raises_when_defaults_in_wrong_order():

    with pytest.raises(TypeError):
        class X(Record):
            foo: str
            bar: int = 3
            baz: str


def test_maybe_namespace_raises_for_missing_abstract_type():
    class X(Record):
        foo: str

    with pytest.raises(KeyError):
        X._maybe_namespace({X._blessed_key: {'ns': 'a.b.c.d.e.f'}},
                           preferred_type=ModelT)


def test_compat_loads_DeprecationWarning():
    class X(Record):
        foo: str

    payload = X('foo').dumps(serializer='json')
    with pytest.warns(DeprecationWarning):
        X.loads(payload, default_serializer='json')


def test_model_overriding_Options_sets_options():

    class X(Record):
        foo: str

        class Options:
            coercions = {'foo': 'bar'}
            foo = 1.345

    with pytest.raises(AttributeError):
        X.Options
    assert X._options.coercions == {'foo': 'bar'}
    assert X._options.foo == 1.345


def test_model_with_custom_hash():
    class X(Record):
        foo: int

        def __hash__(self):
            return self.foo

    assert hash(X(123)) == 123


def test_model_with_custom_eq():
    class X(Record):
        foo: bool

        def __eq__(self, other):
            return True

    assert X(10) == X(20)


def test_Record_comparison():
    class X(Record):
        x: int
        y: int

    assert X(10, 30) > X(8, 24)
    assert X(10, 30) >= X(10, 30)
    assert X(10, 30) < X(10, 40)
    assert X(10, 40) <= X(10, 40)

    with pytest.raises(TypeError):
        X(10) >= object()
    with pytest.raises(TypeError):
        X(10) <= object()
    with pytest.raises(TypeError):
        X(10) < object()
    with pytest.raises(TypeError):
        X(10) > object()
    with pytest.raises(TypeError):
        object() > X(10)
    with pytest.raises(TypeError):
        object() >= X(10)
    with pytest.raises(TypeError):
        object() < X(10)
    with pytest.raises(TypeError):
        object() <= X(10)


def test_maybe_model():

    class X(Record):
        x: int
        y: int

    assert maybe_model('foo') == 'foo'
    assert maybe_model(1) == 1
    assert maybe_model(1.01) == 1.01

    x1 = X(10, 20)
    assert maybe_model(json.loads(x1.dumps(serializer='json'))) == x1


def test_StringField():

    class Moo(Record):
        foo: str = StringField(max_length=10, min_length=3, allow_blank=False)

    too_long_moo = Moo('thequickbrownfoxjumpsoverthelazydog')
    too_short_moo = Moo('xo')
    perfect_moo = Moo('foobar')

    assert perfect_moo.is_valid()
    assert not too_long_moo.is_valid()
    assert not too_short_moo.is_valid()

    assert not perfect_moo.validation_errors
    assert too_long_moo.validation_errors
    assert too_short_moo.validation_errors

    assert too_long_moo.foo == 'thequickbrownfoxjumpsoverthelazydog'
    assert too_short_moo.foo == 'xo'
    assert perfect_moo.foo == 'foobar'

    assert not Moo('').is_valid()
    assert Moo('').validation_errors
    assert 'blank' in str(Moo('').validation_errors[0])


def test_StringField_optional__explicit():

    class Moo(Record):
        foo: str = StringField(max_length=10,
                               min_length=3,
                               required=False,
                               allow_blank=False)

    moo = Moo()
    assert moo.foo is None
    assert moo.is_valid()


def test_StringField_optional__Optional():

    class Moo(Record):
        foo: Optional[str] = StringField(max_length=10,
                                         min_length=3,
                                         allow_blank=False)

    moo = Moo()
    assert moo.foo is None
    assert moo.is_valid()


def test_validation_ensures_types_match():

    class Order(Record, validation=True):
        price: Decimal = DecimalField()
        quantity: int = IntegerField()
        side: str = StringField()
        foo: float = FloatField(required=False, default=3.33)

    # Field descriptors are not considered defaults
    with pytest.raises(TypeError):
        Order()
    with pytest.raises(TypeError):
        Order(price=Decimal('3.13'))
    with pytest.raises(TypeError):
        Order(price=Decimal('3.13'), quantity=30)

    with pytest.raises(ValidationError):
        Order(price='NaN', quantity=2, side='BUY')
    with pytest.raises(ValueError):
        Order(price='3.13', quantity='xyz', side='BUY')

    assert Order(price=Decimal(3.13), quantity=10, side='BUY')

    order = Order(price='3.13', quantity='10', side='BUY', foo='3.33')
    assert order.price == Decimal('3.13')
    assert isinstance(order.price, Decimal)
    assert order.quantity == 10
    assert order.side == 'BUY'
    assert order.foo == 3.33
    assert isinstance(order.foo, float)


def test_Decimal_max_digits():

    class X(Record, validation=True):
        foo: Decimal = DecimalField(max_digits=3)

    with pytest.raises(ValidationError):
        X(foo='3333.3333333333')

    assert X(foo='33.33333333333333333').foo


def test_StringField_trim_whitespace():

    class X(Record, validation=True):
        foo: str = StringField(trim_whitespace=True)

    assert X(foo='  the quick    ').foo == 'the quick'


def test_BytesField():

    class X(Record, validation=True):
        foo: bytes = BytesField(max_length=10, min_length=3)

    assert X(foo='foo').foo == b'foo'


def test_BytesField_trim_whitespace():

    class X(Record, validation=True):
        foo: bytes = BytesField(trim_whitespace=True)

    assert X(foo='  the quick    ').foo == b'the quick'


def test_field_descriptors_may_mix_with_non_defaults():

    class Person(faust.Record, validation=True):
        age: int = IntegerField(min_value=18, max_value=200)
        name: str

    with pytest.raises(ValidationError):
        Person(age=9, name='Robin')

    with pytest.raises(ValidationError):
        Person(age=203, name='Abraham Lincoln')


def test_field_descriptors_throws_type_error():

    class Person(faust.Record):
        age: int
        name: str

    person = Person(age='Batman', name='Robin')
    assert person.validation_errors


def test_implicit_descritor_types():

    class X(Record):
        a: int
        b: float
        c: str
        d: bytes
        e: Decimal

    assert isinstance(X.a, IntegerField)
    assert isinstance(X.b, FloatField)
    assert isinstance(X.c, StringField)
    assert isinstance(X.d, BytesField)
    assert isinstance(X.e, DecimalField)


def test_exclude():

    class X(Record):
        a: str
        b: str
        c: str = StringField(exclude=True)

    x = X('A', 'B', 'C')

    assert X.a.required
    assert X.b.required
    assert X.c.required

    assert not X.a.exclude
    assert not X.b.exclude
    assert X.c.exclude

    assert x.asdict() == {'a': 'A', 'b': 'B'}

    assert 'c' not in x.to_representation()


def test_custom_field_validation():

    class ChoiceField(FieldDescriptor[str]):

        def __init__(self, choices: List[str], **kwargs) -> None:
            self.choices = choices
            # Must pass any custom args to init,
            # so we pass the choices keyword argument also here.
            super().__init__(choices=choices, **kwargs)

        def validate(self, value: str):
            if value not in self.choices:
                choices = ', '.join(self.choices)
                yield self.validation_error(
                    f'Field {self.field} must be one of {choices}')

    class Order(faust.Record, validation=True):
        side: str = ChoiceField(['SELL', 'BUY'])

    with pytest.raises(ValidationError):
        Order(side='LEFT')

    class Order2(faust.Record, validation=True):
        side: str = ChoiceField(['SELL', 'BUY'], required=False)

    assert Order2()


def test_custom_field__internal_errot():

    class XField(FieldDescriptor[str]):

        def prepare_value(self, value, coerce=None):
            if coerce:
                raise RuntimeError()
            return value

    class Foo(Record, coerce=False):
        foo: str = XField()

    f = Foo('foo')
    assert f.validation_errors
    assert 'RuntimeError' in str(f.validation_errors[0])


def test_datetime_does_not_coerce():

    class X(Record, coerce=False):
        d: datetime

    date_string = 'Sat Jan 12 00:44:36 +0000 2019'
    assert X(date_string).d == date_string


def test_datetime_custom_date_parser():

    class X(Record, coerce=True, date_parser=parse_date):
        d: datetime

    date_string = 'Sat Jan 12 00:44:36 +0000 2019'
    assert X.from_data({'d': date_string}).d == parse_date(date_string)


def test_float_does_not_coerce():

    class X(Record, coerce=False):
        f: float
    X.make_final()  # <-- just to test it still works for non-lazy creation

    assert X('3.14').f == '3.14'


def test_payload_with_reserved_keyword():

    class X(Record):
        location: str = StringField(input_name='in')
        foo: str = StringField(required=False, default='BAR',
                               input_name='bar', output_name='foobar')

    with pytest.raises(TypeError):
        X()

    assert X('foo').location == 'foo'
    assert X(location='FOO').location == 'FOO'

    d = X.from_data({'in': 'foo', 'bar': 'bar'})
    assert d.location == 'foo'
    assert d.foo == 'bar'

    assert d.asdict() == {
        'in': 'foo',
        'foobar': 'bar',
    }


class LazyX(Record, lazy_creation=True):
    y: 'EagerY'


class EagerY(Record):
    x: LazyX


def test_lazy_creation():
    assert LazyX._pending_finalizers
    LazyX.make_final()
    assert LazyX._pending_finalizers is None


def test_Sensitive(*, capsys):

    class Foo(Record):
        name: str
        phone_number: Sensitive[str]

    class Bar(Record):
        alias: str
        foo: Foo

    class Baz(Record):
        alias: str
        bar: List[Bar]

    x = Foo(name='Foo', phone_number='631-342-3412')

    assert Foo._options.has_sensitive_fields
    assert Foo._options.has_tagged_fields
    assert not Foo._options.has_secret_fields
    assert not Foo._options.has_personal_fields

    assert 'phone_number' in Foo._options.sensitive_fields
    assert 'foo' not in Foo._options.sensitive_fields

    # Model with related model that has sensitive fields, is also sensitive.
    assert Bar._options.has_sensitive_fields
    assert 'foo' in Bar._options.sensitive_fields
    assert 'alias' not in Bar._options.sensitive_fields

    assert Baz._options.has_sensitive_fields
    assert 'bar' in Baz._options.sensitive_fields
    assert 'alias' not in Baz._options.sensitive_fields

    assert x.name == 'Foo'
    with pytest.raises(SecurityError):
        str(x.phone_number)
    assert x.phone_number.get_value() == '631-342-3412'

    with pytest.raises(SecurityError):
        f'Name={x.name} Phone={x.phone_number}'

    logger.critical('User foo error %s', x.phone_number)
    stderr_content = capsys.readouterr()
    assert 'Logging error' in stderr_content.err
    assert 'SecurityError' in stderr_content.err

    def exclaim(x: str) -> str:
        assert isinstance(x, _FrameLocal)
        return f'{x}!'

    with pytest.raises(SecurityError):
        exclaim(x.phone_number.get_value())

    def upper(x: str) -> str:
        return x.upper()

    with pytest.raises(SecurityError):
        upper(x.phone_number.get_value())

    with allow_protected_vars():
        assert upper(x.phone_number.get_value()) == '631-342-3412'


def test_Secret(*, caplog):

    class Foo(Record):
        name: str
        phone_number: Secret[str]

    class Bar(Record):
        alias: str
        foo: Foo

    class Baz(Record):
        alias: str
        bar: List[Bar]

    x = Foo(name='Foo', phone_number='631-342-3412')

    assert Foo._options.has_secret_fields
    assert not Foo._options.has_sensitive_fields
    assert Foo._options.has_tagged_fields
    assert not Foo._options.has_personal_fields

    assert 'phone_number' in Foo._options.secret_fields
    assert 'foo' not in Foo._options.secret_fields

    # Model with related model that has sensitive fields, is also sensitive.
    assert Bar._options.has_secret_fields
    assert 'foo' in Bar._options.secret_fields
    assert 'alias' not in Bar._options.secret_fields

    assert Baz._options.has_secret_fields
    assert 'bar' in Baz._options.secret_fields
    assert 'alias' not in Baz._options.secret_fields

    assert x.name == 'Foo'
    assert str(x.phone_number) == x.phone_number.mask
    assert x.phone_number.get_value() == '631-342-3412'

    assert (f'Name={x.name} Phone={x.phone_number}' ==
            f'Name={x.name} Phone={x.phone_number.mask}')

    logger.critical('User foo error %s', x.phone_number)

    assert x.phone_number.get_value() not in caplog.text
    assert x.phone_number.mask in caplog.text



import pytest
from faust import Event, Table
from faust.stores.base import SerializedStore, Store
from faust.types import TP
from mode import label
from mode.utils.mocks import Mock


class MyStore(Store):

    def __getitem__(self, key):
        ...

    def __setitem__(self, key, value):
        ...

    def __delitem__(self, key):
        ...

    def __iter__(self):
        ...

    def __len__(self):
        ...

    def apply_changelog_batch(self, *args, **kwargs):
        ...

    def reset_state(self):
        ...


class test_Store:

    @pytest.fixture
    def store(self, *, app):
        return MyStore(
            url='foo://',
            app=app,
            table=Mock(name='table'),
            key_serializer='json',
            value_serializer='json')

    def test_persisted_offset(self, *, store):
        with pytest.raises(NotImplementedError):
            store.persisted_offset(TP('foo', 0))

    def test_set_persisted_offset(self, *, store):
        store.set_persisted_offset(TP('foo', 0), 30303)

    @pytest.mark.asyncio
    async def test_need_active_standby_for(self, *, store):
        assert await store.need_active_standby_for(TP('foo', 0))

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, store):
        await store.on_rebalance(
            Mock(name='table', autospec=Table),
            set(),
            set(),
            set(),
        )

    def test_encode_key(self, *, store):
        assert store._encode_key({'foo': 1}) == b'{"foo": 1}'

    def test_encode_key__cannot_be_None(self, *, store):
        store.key_serializer = 'raw'
        with pytest.raises(TypeError):
            store._encode_key(None)

    def test_encode_value(self, *, store):
        assert store._encode_value({'foo': 1}) == b'{"foo": 1}'

    def test_decode_key(self, *, store):
        assert store._decode_key(b'{"foo": 1}') == {'foo': 1}

    def test_decode_value(self, *, store):
        assert store._decode_value(b'{"foo": 1}') == {'foo': 1}

    def test_repr(self, *, store):
        assert repr(store)

    def test_label(self, *, store):
        assert label(store)


class MySerializedStore(SerializedStore):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keep = {}

    def _get(self, key):
        return self.keep.get(key)

    def _set(self, key, value):
        self.keep[key] = value

    def _del(self, key):
        self.keep.pop(key, None)

    def _iterkeys(self):
        return (k for k in self.keep)

    def _itervalues(self):
        return (self.keep[k] for k in self.keep)

    def _iteritems(self):
        return ((k, self.keep[k]) for k in self.keep)

    def _size(self):
        return len(self.keep)

    def _contains(self, key):
        return key in self.keep

    def _clear(self):
        self.keep.clear()

    def reset_state(self):
        ...

    def _prefix_scan(self, prefix: bytes):
        ...        


class test_SerializedStore:

    @pytest.fixture
    def store(self, *, app):
        return MySerializedStore(
            url='foo://',
            app=app,
            table=Mock(name='table'),
            key_serializer='json',
            value_serializer='json',
        )

    def test_apply_changelog_batch(self, *, store):
        event = Mock(name='event', autospec=Event)
        event.message.key = b'foo'
        event.message.value = b'bar'
        store.apply_changelog_batch([event], to_key=Mock(), to_value=Mock())
        assert store.keep[b'foo'] == b'bar'

    def test_apply_changelog_batch__delete_None_value(self, *, store):
        self.test_apply_changelog_batch(store=store)
        assert store.keep[b'foo'] == b'bar'
        event = Mock(name='event', autospec=Event)
        event.message.key = b'foo'
        event.message.value = None
        store.apply_changelog_batch([event], to_key=Mock(), to_value=Mock())
        with pytest.raises(KeyError):
            store.keep[b'foo']

    def test_apply_changelog_batch__key_is_None(self, *, store):
        event = Mock(name='event', autospec=Event)
        event.message.key = None
        event.message.value = b'bar'
        with pytest.raises(TypeError):
            store.apply_changelog_batch(
                [event], to_key=Mock(), to_value=Mock())

    def test_setitem__getitem__delitem(self, *, store):
        store['foo'] = '303'
        with pytest.raises(KeyError):
            store[{'foo': 2}]
        assert store['foo'] == '303'
        assert len(store) == 1
        assert 'foo' in store
        assert list(iter(store)) == ['foo']
        keys = store.keys()
        assert list(iter(keys)) == ['foo']
        values = store.values()
        assert list(iter(values)) == ['303']
        items = store.items()
        assert list(iter(items)) == [('foo', '303')]
        del(store['foo'])
        assert not len(store)
        store['foo'] = '303'
        store.clear()
        assert not len(store)



"""Models."""
from .base import Model, ModelOptions, maybe_model, registry
from .fields import FieldDescriptor, StringField
from .record import Record

__all__ = [
    'FieldDescriptor',
    'Model',
    'ModelOptions',
    'Record',
    'StringField',
    'maybe_model',
    'registry',
]
